/*
 * Copyright (c) 2016, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*********************************************************************
 * INCLUDES
 */
#include <string.h>

//#define xdc_runtime_Log_DISABLE_ALL 1  // Add to disable logs from this file

#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/knl/Queue.h>
#include <ti/sysbios/BIOS.h>

#include <ti/drivers/PIN.h>
#include <ti/drivers/SPI.h>
#include <ti/drivers/I2C.h>
#include <ti/drivers/i2c/I2CCC26XX.h>

#include <xdc/runtime/Log.h>
#include <xdc/runtime/Diags.h>

// Stack headers
#include <hci_tl.h>
#include <gap.h>
#include <gatt.h>
#include <gapgattserver.h>
#include <gattservapp.h>
#include <osal_snv.h>
#include <gapbondmgr.h>
#include <peripheral.h>
#include <ICallBleAPIMSG.h>
#include <math.h>

#include <devinfoservice.h>
#include "WDMM.h"

#include "util.h"

#include "rf430.h"

// Bluetooth Developer Studio services
#include "../Application_Profiles/WDMM_Service.h"
#include "../Application_Profiles/BQ27426_Fuel_Gauge_Service.h"
#include "../Startup/Board.h"

#define BV(n)               (1 << (n))
#define TR10440_LOW_RESOLUTION					// use LIR2477_HIGH_RESOLUTION definition for TIDA-01014 hardware configuration

/*********************************************************************
 * CONSTANTS
 */

#define WDMM_PERIODIC_EVT_PERIOD			 	1000 	// mSec
#define BATTERY_CHECK_INTERVAL_COUNT          	60		// battery check time inverval = WDMM_PERIODIC_EVT_PERIOD * BATTERY_CHECK_INTERVAL_COUNT
#define APD_TIME_INTERVAL_COUNT					900		// automatic power down time = WDMM_PERIODIC_EVT_PERIOD * APD_TIME_INTERVAL_COUNT

// Advertising interval when device is discoverable (units of 625us, 160=100ms)
#define DEFAULT_ADVERTISING_INTERVAL          	1600

// Limited discoverable mode advertises for 30.72s, and then stops
// General discoverable mode advertises indefinitely
#define DEFAULT_DISCOVERABLE_MODE             	GAP_ADTYPE_FLAGS_GENERAL

// Default pass-code used for pairing.
#define DEFAULT_PASSCODE                     	 000000

// Task configuration
#define WDMM_TASK_PRIORITY                    	3
#define ADC_READ_TASK_PRIORITY                	1
#define PWR_MGMT_TASK_PRIORITY                	2

#define WDMM_TASK_STACK_SIZE                  	600
#define ADC_READ_TASK_STACK_SIZE              	1800
#define PWR_MGMT_TASK_STACK_SIZE              	700

// Internal Events for RTOS application
#define WDMM_STATE_CHANGE_EVT                 	0x0001
#define WDMM_CHAR_CHANGE_EVT                  	0x0002
#define WDMM_PERIODIC_EVT                    	0x0004
#define WDMM_CONN_EVT_END_EVT                 	0x0008
#define ADC_READ_READY_EVT					  	0X0010
#define WDMM_POWER_DOWN_EVT				 	  	0x0020
#define WDMM_POWER_UP_EVT				 	  	0x0040
#define WDMM_BATTERY_REPORT_EVT				  	0x0080
#define WDMM_CAPTIVATE_SENSING_EVT			  	0x0100
#define WDMM_NFC_EVT						  	0x0200

#define SNV_ID_VOLTAGE_50mV					  	0x80
#define SNV_ID_VOLTAGE_500mV				  	0x81
#define SNV_ID_VOLTAGE_5V					  	0x82
#define SNV_ID_VOLTAGE_50V					  	0x83
#define SNV_ID_CURRENT_500uA				  	0x84
#define SNV_ID_CURRENT_50mA				  	  	0x85

#define BLE_ADDRESS_START_INDEX					84	// NDEF Buffer Index Starts @ 2+7+17+4 = 30. BLE Address Starts at 30+6+9+37+2 = 84
#define TK_START_INDEX							95	// NDEF Buffer Index Starts @ 2+7+17+4 = 30. TK Starts at 30+6+9+37+8+3+2 = 95
#define I2C_TIMEOUT 							2500

// BQ27426 Set-Up
#define BQ27426_DEVICE_ADDRESS 					0x55

#define Ra_Tables_SubClass_ID					0x5900
#define Current_Thresholds_SubClass_ID			0x5100
#define State_SubClass_ID						0x5200
#define IT_Cfg_0_SubClass_ID					0x5000
#define IT_Cfg_1_SubClass_ID					0x5901
#define IT_Cfg_2_SubClass_ID					0x5902
#define Chem_Data_SubClass_ID					0x6D00
#define Charge_Termination_SubClass_ID			0x2400
#define Discharge_SubClass_ID					0x3100
#define Registers_SubClass_ID					0x4000
#define Power_SubClass_ID						0x4400
#define OCVa_0_SubClass_ID						0x5300
#define OCVa_1_SubClass_ID						0x5301
#define OCVb_0_SubClass_ID						0x5400
#define OCVb_1_SubClass_ID						0x5401
#define Resistance_Low_Temp_SubClass_ID			0x5500
#define Resistance_High_Temp_SubClass_ID		0x6C00

#ifdef TR10440_HIGH_RESOLUTION
static const uint8_t Ra_Tables[32] = {0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x02, 0x00, 0x03, 0x00, 0x00};
static const uint8_t Current_Thresholds[32] = {0x03, 0xA5, 0x00, 0xE9, 0x05, 0x78, 0x00, 0x3C, 0x3C, 0x01, 0xB3, 0xB3, 0x01, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t State[32] = {0x40, 0x00, 0x00, 0x00, 0x00, 0x81, 0x1B, 0x58, 0x65, 0x2C, 0x0C, 0x80, 0x00, 0xC8, 0x00, 0x32, 0x00, 0x14, 0x03, 0xE8, 0x01, 0x01, 0x75, 0x00, 0x02, 0xFF, 0xCE, 0xFF, 0xCE, 0x00, 0x01, 0x00};
static const uint8_t IT_Cfg_0[32] = {0x01, 0xF4, 0x00, 0x1E, 0xC8, 0x14, 0x08, 0x00, 0x3C, 0x0E, 0x10, 0x00, 0x0A, 0x46, 0x05, 0x14, 0x05, 0x0F, 0x03, 0x20, 0x7F, 0xFF, 0x00, 0xF0, 0x46, 0x50, 0x18, 0x01, 0x90, 0x00, 0x64, 0x19};
static const uint8_t IT_Cfg_1[32] = {0xDC, 0x5C, 0x60, 0x00, 0x7D, 0x00, 0x04, 0x03, 0x19, 0x25, 0x0F, 0x14, 0x0A, 0x78, 0x60, 0x28, 0x01, 0xF4, 0x00, 0x00, 0x00, 0x00, 0x43, 0x80, 0x04, 0x01, 0x14, 0x00, 0x08, 0x0B, 0xB8, 0x01};
static const uint8_t IT_Cfg_2[32] = {0x2C, 0x0A, 0x01, 0x0A, 0x00, 0x00, 0x00, 0xC8, 0x00, 0x64, 0x02, 0x00, 0x00, 0x00, 0x00, 0x07, 0xD0, 0x01, 0x03, 0x5A, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Chem_Data[32] = {0x1B, 0x58, 0x0E, 0xDD, 0x0E, 0xA2, 0x10, 0x68, 0x10, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Charge_Termination[32] = {0x01, 0xF4, 0x28, 0x63, 0x5F, 0xFF, 0x62, 0x00, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Discharge[32] = {0x0A, 0x0F, 0x02, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Registers[32] = {0x64, 0x7C, 0x0F, 0x9F, 0x23, 0x00, 0x00, 0x14, 0x04, 0x00, 0x09, 0x04, 0x26, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Power[32] = {0x00, 0x32, 0x01, 0xC2, 0x30, 0x00, 0x03, 0x08, 0x98, 0x01, 0x00, 0x3C, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t OCVa_0[32] = {0x31, 0x42, 0x11, 0x2A, 0xDE, 0xDF, 0xDF, 0xE1, 0xE0, 0xE1, 0xE2, 0xE3, 0xE1, 0xE6, 0xE7, 0xE0, 0xE3, 0xEA, 0xE9, 0xEE, 0xE8, 0xDF, 0xF0, 0xEC, 0xF4, 0xF5, 0xF5, 0xF7, 0xF9, 0xF9, 0xFE, 0xFB};
static const uint8_t OCVa_1[32] = {0xFF, 0xF6, 0xEF, 0xEE, 0xE9, 0xF5, 0xF9, 0xFA, 0xF1, 0xD8, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t OCVb_0[32] = {0xFC, 0xC6, 0x04, 0x0C, 0x07, 0x0D, 0xFD, 0x05, 0x05, 0x00, 0xFD, 0x0C, 0xFD, 0xF9, 0x21, 0x13, 0x0A, 0x02, 0x00, 0xF7, 0x11, 0xD7, 0xFA, 0xF6, 0xF6, 0x07, 0xFD, 0xFB, 0xFE, 0xE2, 0xF0, 0xE0};
static const uint8_t OCVb_1[32] = {0xEF, 0x16, 0x02, 0xD3, 0x09, 0x26, 0xFB, 0x64, 0xA1, 0xC2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Resistance_Low_Temp[32] = {0xFF, 0x2F, 0x00, 0x02, 0xFB, 0x07, 0x1D, 0xDD, 0x19, 0xB7, 0x03, 0xFA, 0x2F, 0xDB, 0x26, 0xF1, 0x4D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Resistance_High_Temp[32] = {0xFE, 0x4F, 0x02, 0x03, 0x06, 0x0C, 0xFF, 0xF2, 0xFF, 0xF7, 0xFC, 0x00, 0xF9, 0xFC, 0xF8, 0x06, 0xBF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
#endif

#ifdef TR10440_LOW_RESOLUTION
static const uint8_t State[32] = {0x40, 0x00, 0x00, 0x00, 0x00, 0x81, 0x01, 0x5E, 0x05, 0x0F, 0x0C, 0x80, 0x00, 0xC8, 0x00, 0x32, 0x00, 0x14, 0x03, 0xE8, 0x01, 0x01, 0x75, 0x00, 0x02, 0xFF, 0xCE, 0xFF, 0xCE, 0x00, 0x01, 0x00};
static const uint8_t Current_Thresholds[32] = {0x03, 0xA5, 0x00, 0xE9, 0x05, 0x78, 0x00, 0x3C, 0x3C, 0x01, 0xB3, 0xB3, 0x01, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
#endif

#ifdef LIR2477_HIGH_RESOLUTION
static const uint8_t Ra_Tables[32] = {0x00, 0x24, 0x00, 0x00, 0x00, 0x09, 0x00, 0x14, 0x00, 0x14, 0x00, 0x13, 0x00, 0x12, 0x00, 0x11, 0x00, 0x10, 0x00, 0x0F, 0x00, 0x0F, 0x00, 0x10, 0x00, 0x12, 0x00, 0x1F, 0x00, 0x2F, 0x00, 0x00};
static const uint8_t Current_Thresholds[32] = {0x01, 0x90, 0x00, 0x64, 0x02, 0x58, 0x00, 0x3C, 0x3C, 0x01, 0xB3, 0xB3, 0x01, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t State[32] = {0x4C, 0xCD, 0x00, 0x00, 0x00, 0x81, 0x0B, 0xB8, 0x2B, 0x5C, 0x0C, 0x80, 0x00, 0xC8, 0x00, 0x32, 0x00, 0x14, 0x03, 0xE8, 0x01, 0x00, 0x78, 0x00, 0x02, 0xFF, 0xCA, 0xFF, 0xCD, 0x00, 0x01, 0x00};
static const uint8_t IT_Cfg_0[32] = {0x01, 0xF4, 0x00, 0x1E, 0xC8, 0x14, 0x08, 0x00, 0x3C, 0x0E, 0x10, 0x00, 0x0A, 0x46, 0x05, 0x14, 0x05, 0x0F, 0x03, 0x20, 0x7F, 0xFF, 0x00, 0xF0, 0x46, 0x50, 0x18, 0x01, 0x90, 0x00, 0x64, 0x19};
static const uint8_t IT_Cfg_1[32] = {0xDC, 0x5C, 0x60, 0x00, 0x7D, 0x00, 0x04, 0x03, 0x19, 0x25, 0x0F, 0x14, 0x0A, 0x78, 0x60, 0x28, 0x01, 0xF4, 0x00, 0x00, 0x00, 0x00, 0x43, 0x80, 0x04, 0x01, 0x14, 0x00, 0x08, 0x0B, 0xB8, 0x01};
static const uint8_t IT_Cfg_2[32] = {0x2C, 0x0A, 0x01, 0x0A, 0x00, 0x00, 0x00, 0xC8, 0x00, 0x64, 0x02, 0x00, 0x00, 0x00, 0x00, 0x07, 0xD0, 0x01, 0x03, 0x5A, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Chem_Data[32] = {0x14, 0xEF, 0x0E, 0xDD, 0x0E, 0xA2, 0x10, 0x62, 0x10, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Charge_Termination[32] = {0x01, 0xF4, 0x28, 0x63, 0x5F, 0xFF, 0x62, 0x00, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Discharge[32] = {0x0A, 0x0F, 0x02, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Registers[32] = {0x64, 0x78, 0x4F, 0x9F, 0x23, 0x00, 0x00, 0x14, 0x04, 0x00, 0x09, 0x04, 0x26, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Power[32] = {0x00, 0x32, 0x01, 0xC2, 0x30, 0x00, 0x03, 0x08, 0x98, 0x01, 0x00, 0x3C, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t OCVa_0[32] = {0x31, 0x42, 0x11, 0x2A, 0xDE, 0xDF, 0xDF, 0xE1, 0xE0, 0xE1, 0xE2, 0xE3, 0xE1, 0xE6, 0xE7, 0xE0, 0xE3, 0xEA, 0xE9, 0xEE, 0xE8, 0xDF, 0xF0, 0xEC, 0xF4, 0xF5, 0xF5, 0xF7, 0xF9, 0xF9, 0xFE, 0xFB};
static const uint8_t OCVa_1[32] = {0xFF, 0xF6, 0xEF, 0xEE, 0xE9, 0xF5, 0xF9, 0xFA, 0xF1, 0xD8, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t OCVb_0[32] = {0xFC, 0xC6, 0x04, 0x0C, 0x07, 0x0D, 0xFD, 0x05, 0x05, 0x00, 0xFD, 0x0C, 0xFD, 0xF9, 0x21, 0x13, 0x0A, 0x02, 0x00, 0xF7, 0x11, 0xD7, 0xFA, 0xF6, 0xF6, 0x07, 0xFD, 0xFB, 0xFE, 0xE2, 0xF0, 0xE0};
static const uint8_t OCVb_1[32] = {0xEF, 0x16, 0x02, 0xD3, 0x09, 0x26, 0xFB, 0x64, 0xA1, 0xC2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Resistance_Low_Temp[32] = {0xFF, 0x2F, 0x00, 0x02, 0xFB, 0x07, 0x1D, 0xDD, 0x19, 0xB7, 0x03, 0xFA, 0x2F, 0xDB, 0x26, 0xF1, 0x4D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Resistance_High_Temp[32] = {0xFE, 0x4F, 0x02, 0x03, 0x06, 0x0C, 0xFF, 0xF2, 0xFF, 0xF7, 0xFC, 0x00, 0xF9, 0xFC, 0xF8, 0x06, 0xBF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
#endif

#ifdef LIR2477_LOW_RESOLUTION
static const uint8_t Ra_Tables[32] = {0x00, 0x24, 0x00, 0x00, 0x00, 0x09, 0x00, 0x14, 0x00, 0x14, 0x00, 0x13, 0x00, 0x12, 0x00, 0x11, 0x00, 0x10, 0x00, 0x0F, 0x00, 0x0F, 0x00, 0x10, 0x00, 0x12, 0x00, 0x1F, 0x00, 0x2F, 0x00, 0x00};
static const uint8_t Current_Thresholds[32] = {0x01, 0x90, 0x00, 0x64, 0x02, 0x58, 0x00, 0x3C, 0x3C, 0x01, 0xB3, 0xB3, 0x01, 0x90, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t State[32] = {0x4C, 0xCD, 0x00, 0x00, 0x00, 0x81, 0x00, 0x96, 0x02, 0x2B, 0x0C, 0x80, 0x00, 0xC8, 0x00, 0x32, 0x00, 0x14, 0x03, 0xE8, 0x01, 0x00, 0x78, 0x00, 0x02, 0xFF, 0xCA, 0xFF, 0xCD, 0x00, 0x01, 0x00};
static const uint8_t IT_Cfg_0[32] = {0x01, 0xF4, 0x00, 0x1E, 0xC8, 0x14, 0x08, 0x00, 0x3C, 0x0E, 0x10, 0x00, 0x0A, 0x46, 0x05, 0x14, 0x05, 0x0F, 0x03, 0x20, 0x7F, 0xFF, 0x00, 0xF0, 0x46, 0x50, 0x18, 0x01, 0x90, 0x00, 0x64, 0x19};
static const uint8_t IT_Cfg_1[32] = {0xDC, 0x5C, 0x60, 0x00, 0x7D, 0x00, 0x04, 0x03, 0x19, 0x25, 0x0F, 0x14, 0x0A, 0x78, 0x60, 0x28, 0x01, 0xF4, 0x00, 0x00, 0x00, 0x00, 0x43, 0x80, 0x04, 0x01, 0x14, 0x00, 0x08, 0x0B, 0xB8, 0x01};
static const uint8_t IT_Cfg_2[32] = {0x2C, 0x0A, 0x01, 0x0A, 0x00, 0x00, 0x00, 0xC8, 0x00, 0x64, 0x02, 0x00, 0x00, 0x00, 0x00, 0x07, 0xD0, 0x01, 0x03, 0x5A, 0x14, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Chem_Data[32] = {0x14, 0xEF, 0x0E, 0xDD, 0x0E, 0xA2, 0x10, 0x62, 0x10, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Charge_Termination[32] = {0x00, 0x19, 0x28, 0x63, 0x5F, 0xFF, 0x62, 0x00, 0x32, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Discharge[32] = {0x0A, 0x0F, 0x02, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Registers[32] = {0x64, 0x78, 0x4F, 0x9F, 0x23, 0x00, 0x00, 0x14, 0x04, 0x00, 0x09, 0x04, 0x26, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Power[32] = {0x00, 0x32, 0x01, 0xC2, 0x30, 0x00, 0x03, 0x08, 0x98, 0x01, 0x00, 0x3C, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t OCVa_0[32] = {0x31, 0x42, 0x11, 0x2A, 0xDE, 0xDF, 0xDF, 0xE1, 0xE0, 0xE1, 0xE2, 0xE3, 0xE1, 0xE6, 0xE7, 0xE0, 0xE3, 0xEA, 0xE9, 0xEE, 0xE8, 0xDF, 0xF0, 0xEC, 0xF4, 0xF5, 0xF5, 0xF7, 0xF9, 0xF9, 0xFE, 0xFB};
static const uint8_t OCVa_1[32] = {0xFF, 0xF6, 0xEF, 0xEE, 0xE9, 0xF5, 0xF9, 0xFA, 0xF1, 0xD8, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t OCVb_0[32] = {0xFC, 0xC6, 0x04, 0x0C, 0x07, 0x0D, 0xFD, 0x05, 0x05, 0x00, 0xFD, 0x0C, 0xFD, 0xF9, 0x21, 0x13, 0x0A, 0x02, 0x00, 0xF7, 0x11, 0xD7, 0xFA, 0xF6, 0xF6, 0x07, 0xFD, 0xFB, 0xFE, 0xE2, 0xF0, 0xE0};
static const uint8_t OCVb_1[32] = {0xEF, 0x16, 0x02, 0xD3, 0x09, 0x26, 0xFB, 0x64, 0xA1, 0xC2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Resistance_Low_Temp[32] = {0xFF, 0x2F, 0x00, 0x02, 0xFB, 0x07, 0x1D, 0xDD, 0x19, 0xB7, 0x03, 0xFA, 0x2F, 0xDB, 0x26, 0xF1, 0x4D, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
static const uint8_t Resistance_High_Temp[32] = {0xFE, 0x4F, 0x02, 0x03, 0x06, 0x0C, 0xFF, 0xF2, 0xFF, 0xF7, 0xFC, 0x00, 0xF9, 0xFC, 0xF8, 0x06, 0xBF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
#endif
/*********************************************************************
 * TYPEDEFS
 */
// Types of messages that can be sent to the user application task from other
// tasks or interrupts. Note: Messages from BLE Stack are sent differently.
typedef enum
{
  APP_MSG_SERVICE_WRITE = 0,   /* A characteristic value has been written     */
  APP_MSG_SERVICE_CFG,         /* A characteristic configuration has changed  */
  APP_MSG_UPDATE_CHARVAL,      /* Request from ourselves to update a value    */
  APP_MSG_GAP_STATE_CHANGE,    /* The GAP / connection state has changed      */
  APP_MSG_SWITCH_DEBOUNCED,    /* A switch has been debounced with new value  */
  APP_MSG_SEND_PASSCODE,       /* A pass-code/PIN is requested during pairing */
} app_msg_types_t;

typedef struct
{
  float	 gain;
  uint32_t dc_offset;
  uint32_t ac_offset;
} cal_constants_t;

// Struct for messages sent to the application task
typedef struct
{
  Queue_Elem       _elem;
  app_msg_types_t  type;
  uint8_t          pdu[];
} app_msg_t;

// Struct for messages about characteristic data
typedef struct
{
  uint16_t svcUUID; // UUID of the service
  uint16_t dataLen; //
  uint8_t  paramID; // Index of the characteristic
  uint8_t  data[];  // Flexible array member, extended to malloc - sizeof(.)
} char_data_t;

// Struct for message about sending/requesting passcode from peer.
typedef struct
{
  uint16_t connHandle;
  uint8_t  uiInputs;
  uint8_t  uiOutputs;
} passcode_req_t;

// Struct for message about switch state
typedef struct
{
  PIN_Id   pinId;
  uint8_t  state;
} switch_state_t;

/*********************************************************************
 * LOCAL VARIABLES
 */
static uint8_t NDEF_Application_Data[142] = {
		// Total Buffer Size = 112 (NDEF Length) + 7 (NDEF App Name) + 17 (CC) + 4 (NDEF File ID/Length) + 2 I2C Address = 133

		// I2C Write Address
		0x00, 0x00,

		// NDEF Tag Application Name (7 bytes)
		0xD2, 0x76, 0x00, 0x00, 0x85, 0x01, 0x01,

		// Capability Container (CC) Contents (17 Bytes)
		0xE1, 0x03,
		0x00, 0x0F, 		// CCLEN
		0x20, 				// Mapping version 2.0
		0x00, 0xF9, 		// MLe (49 bytes); Maximum R-APDU data size
		0x00, 0XF6, 		// MLc (52 bytes); Maximum C-APDU data size
		0x04, 				// Tag, File Control TLV (4 = NDEF file)
		0x06, 				// Length, File Control TLV (6 = 6 bytes of data for this tag)
		0xE1, 0x04, 		// File ID
		0x0B, 0xDF,			// Max NDEF Size
		0x00,				// Read Privileges
		0x00,				// Write Privileges

		//  NDEF File ID + Length (4 bytes)
		0xE1, 0x04, 		// FILE ID
		0x00, 0x70, 		// Length of File = 6+9+37+8+3+18+4+27 = 112 -> 0x67

		// Handover Select (6 bytes)
		0x91, 				// NDEF Record Holder
		0x02, 				// Record Type Length
		0x0A, 				// Record Type Length
		0x48, 0x73, 		// Record Type "hs"
		0x12, 				// Version number 1.2

		0xD1,
		0x02,
		0x04,
		0x61, 0x63,
		0x01,
		0x01,
		0x30,
		0x00,

		// Record Type Name = application/vnd.bluetooth.le.oob (37 bytes)
		0x5A, 				// NDEF Record Header (TNF = 010b)
		0x20, 				// Record Type Name Length = 32 bytes
		0x3C, 				// Payload Length (Starts after Payload ID) = 8+3+18+4+27 = 60 -> 0x3C
		0x01, 				// Payload ID Length = 1
		0x61, 0x70, 0x70, 0x6C, 0x69, 0x63, 0x61, 0x74, // Record Type NAme
		0x69, 0x6F, 0x6E, 0x2F, 0x76, 0x6E, 0x64, 0x2E,
		0x62, 0x6C, 0x75, 0x65, 0x74, 0x6F, 0x6F, 0x74,
		0x68, 0x2E, 0x6C, 0x65, 0x2E, 0x6F, 0x6F, 0x62,
		0x30, 				// Payload ID = 0

		// BLE Pairing Information
		// Device Addess Information (8 bytes)
		0x07, 				// BLE Device Address length = 6
		0x1B,				// Data Type[1] = BLE Device Address
		0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,			   	// BLE MAC ADDRESS - Make this dynamic
		// LE Role Information  (3 bytes)
		0x02, 				// LE Role Length
		0x1C, 				// Data Type[1] = LE Role
		0x00, 				// LE role = Peripheral only
		// Security (18 Bytes)
		0x11,				// Security Key Length
		0x10, 				// Data Type[1] = Security Manager TK Value
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,	// Security TK value - Must Pull from OOB layer of CC26xx
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,

		// Optional Data Section
		// Device Appearance (4 bytes)
		0x03, 				// Appearance length
		0x19, 				// Data Type[1] = Apperance
		0x01, 0x80, 		// Remote Control
		// Device Name (27 bytes)
		0x1A, 				// Complete Local Name length
		0x09, 				// Data Type[1] = Complete Local Name
		0x4e, 0x46, 0x43, 0x20, 0x50, 0x61, 0x69, 0x72, // Name = NFC Paired BLE Peripheral
		0x65, 0x64, 0x20, 0x42, 0x4c, 0x45, 0x20, 0x50,
		0x65, 0x72, 0x69, 0x70, 0x68, 0x65, 0x72, 0x61,
		0x6c
		// End of NDEF

		// References:
		// [1] Generic Access Profile (GAP) - EIR Data Type, Advertising Data Type (AD Type) and OOB Data Type Definitions
		// https://www.bluetooth.org/en-us/specification/assigned-numbers/generic-access-profile
};
// Entity ID globally used to check for source and/or destination of messages
static ICall_EntityID selfEntity;

// Semaphore used for RF430 NFC
static Semaphore_Struct mutex;

// Semaphore globally used to post events to the application thread
static ICall_Semaphore sem;

// Semaphore used to wait for SPI Transaction event
static Semaphore_Struct spiTaskAlert;

// Semaphore used to wait for Power Up event
static Semaphore_Struct pwr_mgmt_TaskAlert;

// Queue object used for application messages.
static Queue_Struct applicationMsgQ;
static Queue_Handle hApplicationMsgQ;

// WDMM Task configuration
Task_Struct wdmmTask;
Char wdmmTaskStack[WDMM_TASK_STACK_SIZE];

// ADC Read Task configuration
Task_Struct adcreadTask;
Char ADCReadTaskStack[ADC_READ_TASK_STACK_SIZE];

// Power Management Task configuration
Task_Struct pwr_mgmtTask;
Char PwrMgmtTaskStack[PWR_MGMT_TASK_STACK_SIZE];

// events flag for internal application events.
static uint16_t events;

// GAP - SCAN RSP data (max size = 31 bytes)
static uint8_t scanRspData[] =
{
  // No scan response data provided.
  0x00 // Placeholder to keep the compiler happy.
};

// GAP - Advertisement data (max size = 31 bytes, though this is
// best kept short to conserve power while advertisting)
static uint8_t advertData[] =
{
  // Flags; this sets the device to use limited discoverable
  // mode (advertises for 30 seconds at a time) or general
  // discoverable mode (advertises indefinitely), depending
  // on the DEFAULT_DISCOVERY_MODE define.
  0x02,   // length of this data
  GAP_ADTYPE_FLAGS,
  DEFAULT_DISCOVERABLE_MODE | GAP_ADTYPE_FLAGS_BREDR_NOT_SUPPORTED,

  // complete name
  5,
  GAP_ADTYPE_LOCAL_NAME_COMPLETE,
 'W','D','M','M',
};

// GAP GATT Attributes
static uint8_t attDeviceName[GAP_DEVICE_NAME_LEN] = "TIDA-01012 WDMM";

// Globals used for ATT Response retransmission
static gattMsgEvent_t *pAttRsp = NULL;
static uint8_t rspTxRetry = 0;

/* Pin driver handles */
static PIN_Handle switchPinHandle;
static PIN_Handle afe_controlPinHandle;
static PIN_Handle rf430_controlPinHandle;
static PIN_Handle power_controlPinHandle;
static SPI_Handle spi_adc;
static I2C_Handle i2c;

/* Global memory storage for a PIN_Config table */
static PIN_State switchPinState;
static PIN_State afe_controlPinState;
static PIN_State power_controlPinState;
static PIN_State rf430_controlPinState;

/*
 * Initial afe_control pin configuration table
 */
PIN_Config afe_controlPinTable[] = {
		VOLTAGE_RANGE_SELECT_1 	| PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
		VOLTAGE_RANGE_SELECT_0	| PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
		CURRENT_RANGE_SELECT   	| PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
		AFE_GAIN_SELECT   		| PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
		CURRENT_SHUNT_ISOLATE  	| PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
		ADS8885_CNVST			| PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
		PIN_TERMINATE
};

/*
 * Application switch pin configuration table:
 */
PIN_Config switchPinTable[] = {
    VOLTAGE_RANGE_SWITCH | PIN_INPUT_EN | PIN_IRQ_BOTHEDGES,
    MEASUREMENT_MODE_SWITCH | PIN_INPUT_EN | PIN_IRQ_BOTHEDGES,
    PIN_TERMINATE
};

/*
* Application RF430 configuration table:
*/
PIN_Config rf430_controlPinTable[] = {
		RF430_IRQ  		| PIN_INPUT_EN | PIN_IRQ_DIS,
		RF430_nRST  	| PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
		RF430_VCORE  	| PIN_INPUT_EN | PIN_IRQ_POSEDGE,
		PIN_TERMINATE
};

/*
* Application power pin configuration table:
*/
PIN_Config power_controlPinTable[] = {
		AFE_DCDC_ENABLE | PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
		AFE_LDO_ENABLE  | PIN_GPIO_OUTPUT_EN | PIN_GPIO_LOW | PIN_PUSHPULL | PIN_DRVSTR_MAX,
		CC2640_WAKE_UP  | PIN_INPUT_EN | PIN_IRQ_POSEDGE,
		MSP430_RESET  	| PIN_GPIO_OUTPUT_EN | PIN_OPENDRAIN,
		ENABLE_BQ27426  | PIN_GPIO_OUTPUT_EN | PIN_OPENDRAIN,
		PIN_TERMINATE
};

// Clock objects for debouncing the switch
static Clock_Struct switch0DebounceClock;
static Clock_Struct switch1DebounceClock;

// State of the switch
static uint8_t switch0State = 0;
static uint8_t switch1State = 0;

// wdmm variables variables
static int32_t send_dc_value;
static int64_t send_ac_value;
volatile static int32_t battery_state_update_count=0;
volatile static int32_t apd_count=0;
static int8_t captivate_active=1;
static Bool wdmm_connected=FALSE;
static Bool nfc_detected=FALSE;
static Bool expedite_power_down=FALSE;

static int16_t temperature, battery_voltage, nominal_available_capacity, full_available_capacity, remaining_capacity = 0;
static int16_t full_charge_capacity, average_current, average_power, state_of_charge = 0;
static int16_t state_of_health, bq27426_flags = 0;
//static int16_t internal_temperature, remaining_capacity_unfiltered, state_of_charge_unfiltered, full_charge_capacity_filtered, full_charge_capacity_unfiltered, remaining_capacity_filtered;

//static Bool temp_adc_on=1;

// Clock instances for internal periodic events.
static Clock_Struct periodicClock;

/*********************************************************************
 * LOCAL FUNCTIONS
 */

static void BLE_init( void );
static void BLE_WDMM_taskFxn(UArg a0, UArg a1);
static void WDMM_adcreadTaskFxn(UArg a0, UArg a1);
static void WDMM_pwr_mgmtTaskFxn(UArg a0, UArg a1);

static void user_processApplicationMessage(app_msg_t *pMsg);
static uint8_t WDMM_processStackMsg(ICall_Hdr *pMsg);
static uint8_t WDMM_processGATTMsg(gattMsgEvent_t *pMsg);

static void WDMM_sendAttRsp(void);
static uint8_t WDMM_processGATTMsg(gattMsgEvent_t *pMsg);
static void WDMM_freeAttRsp(uint8_t status);

static void user_processGapStateChangeEvt(gaprole_States_t newState);
static void user_gapStateChangeCB(gaprole_States_t newState);
static void user_gapBondMgr_passcodeCB(uint8_t *deviceAddr, uint16_t connHandle,
                                       uint8_t uiInputs, uint8_t uiOutputs);
static void user_gapBondMgr_pairStateCB(uint16_t connHandle, uint8_t state,
                                        uint8_t status);
static void switchDebounceSwiFxn(UArg switchId);
static void user_handleSwitchChange(switch_state_t *pState);

// Generic callback handlers for value changes in services.
static void user_service_ValueChangeCB( uint16_t connHandle, uint16_t svcUuid, uint8_t paramID, uint8_t *pValue, uint16_t len );

// Task context handlers for generated services.
static void user_WdmmService_ValueChangeHandler(char_data_t *pCharData);
static void user_Bq27426FuelGaugeService_ValueChangeHandler(char_data_t *pCharData);

// Task handler for sending notifications.

// Utility functions
static void user_enqueueRawAppMsg(app_msg_types_t appMsgType, uint8_t *pData, uint16_t len );
static void user_enqueueCharDataMsg(app_msg_types_t appMsgType, uint16_t connHandle,
                                    uint16_t serviceUUID, uint8_t paramID,
                                    uint8_t *pValue, uint16_t len);
static void switchCallbackFxn(PIN_Handle handle, PIN_Id pinId);

static void SetVoltageRange(char);
static void SetCurrentRange(char);
static void wdmm_power_down();
static BQ27426_read(uint8_t nRegister, uint8_t *pData, uint8_t nLength);
static BQ27426_write(uint8_t *pData, uint8_t nLength);
static uint16_t BQ27426_cmd_read(uint8_t nCmd);
static void BQ27426_cfg_update_mode();
static void BQ27426_soft_reset();
static void BQ27426_reset();
static void BQ27426_configure();
static uint8_t BQ27426_parameter_update(uint8_t nregister, uint8_t byte, uint8_t checksum);
static BQ27426_subclass_update(uint16_t SubClass_ID, const uint8_t *pData);

/*********************************************************************
 * PROFILE CALLBACKS
 */

// GAP Role Callbacks
static gapRolesCBs_t user_gapRoleCBs =
{
  user_gapStateChangeCB     // Profile State Change Callbacks
};

// GAP Bond Manager Callbacks
static gapBondCBs_t user_bondMgrCBs =
{
  user_gapBondMgr_passcodeCB, // Passcode callback
  user_gapBondMgr_pairStateCB // Pairing / Bonding state Callback
};

/*
 * Callbacks in the user application for events originating from BLE services.
 */
// Wdmm Service callback handler.
// The type Data_ServiceCBs_t is defined in Data_Service.h
static WdmmServiceCBs_t user_Wdmm_ServiceCBs =
{
  .pfnChangeCb    = user_service_ValueChangeCB, // Characteristic value change callback handler
  .pfnCfgChangeCb = NULL, // No notification-/indication enabled chars in Wdmm Service
};

static Bq27426FuelGaugeServiceCBs_t user_Bq27426_ServiceCBs =
{
  .pfnChangeCb    = user_service_ValueChangeCB, // Characteristic value change callback handler
  .pfnCfgChangeCb = NULL, // No notification-/indication enabled chars in Wdmm Service
};

static void transferCallback(SPI_Handle handle, SPI_Transaction *transaction)
{
      Semaphore_post(Semaphore_handle(&spiTaskAlert));
}

void wdmmPeriodicTimerAlertCallback(UArg arg) {

	Semaphore_post(Semaphore_handle(&pwr_mgmt_TaskAlert));

} // wdmmPeriodicTimerTaskAlertCallback

static void powerUpCallbackFxn(PIN_Handle handle, PIN_Id pinId)
{
      PIN_setConfig(handle, PIN_BM_IRQ, pinId | PIN_IRQ_DIS);
      apd_count=0;
      events |= WDMM_CAPTIVATE_SENSING_EVT;
      if(!wdmm_connected)
    	  Semaphore_post(Semaphore_handle(&pwr_mgmt_TaskAlert));
}

static void rf430CallbackFxn(PIN_Handle handle, PIN_Id pinId)
{
		if(!wdmm_connected)
		{
    	  nfc_detected = TRUE;
       	  Semaphore_post(Semaphore_handle(&pwr_mgmt_TaskAlert));
		}
}
/*********************************************************************
 * PUBLIC FUNCTIONS
 */

/*
 * @brief   Task creation function for the user task.
 *
 * @param   None.
 *
 * @return  None.
 */
void WDMM_createTask(void)
{
		Task_Params taskParams;
		Task_Params ADCReadTaskParams;		//r
		Task_Params PwrMgmtTaskParams;

		// Configure WDMM task
		Task_Params_init(&taskParams);
		taskParams.stack = wdmmTaskStack;
		taskParams.stackSize = WDMM_TASK_STACK_SIZE;
		taskParams.priority = WDMM_TASK_PRIORITY;
		Task_construct(&wdmmTask, BLE_WDMM_taskFxn, &taskParams, NULL);

		// Configure ADC Read task
		Task_Params_init(&ADCReadTaskParams);
		ADCReadTaskParams.stack = ADCReadTaskStack;
		ADCReadTaskParams.stackSize = ADC_READ_TASK_STACK_SIZE;
		ADCReadTaskParams.priority = ADC_READ_TASK_PRIORITY;
		Task_construct(&adcreadTask, WDMM_adcreadTaskFxn, &ADCReadTaskParams, NULL);

		// Configure Power Management task
		Task_Params_init(&PwrMgmtTaskParams);
		PwrMgmtTaskParams.stack = PwrMgmtTaskStack;
		PwrMgmtTaskParams.stackSize = PWR_MGMT_TASK_STACK_SIZE;
		PwrMgmtTaskParams.priority = PWR_MGMT_TASK_PRIORITY;
		Task_construct(&pwr_mgmtTask, WDMM_pwr_mgmtTaskFxn, &PwrMgmtTaskParams, NULL);

		// Create the semaphore used to wait for SPI callback events
		Semaphore_Params spi_semParams;
		Semaphore_Params_init(&spi_semParams);
		spi_semParams.mode = Semaphore_Mode_BINARY;
		Semaphore_construct(&spiTaskAlert, 0, &spi_semParams);

		// Create the semaphore used to wait for power mgmt callback events
		Semaphore_Params pwr_mgmt_semParams;
		Semaphore_Params_init(&pwr_mgmt_semParams);
		pwr_mgmt_semParams.mode = Semaphore_Mode_BINARY;
		Semaphore_construct(&pwr_mgmt_TaskAlert, 0, &pwr_mgmt_semParams);

		// Create semphore for RF430 NFC
		Semaphore_Params semParamsMutex;

		// Create protection semaphore
		Semaphore_Params_init(&semParamsMutex);
		semParamsMutex.mode = Semaphore_Mode_BINARY;
		Semaphore_construct(&mutex, 1, &semParamsMutex);
}

static void BLE_WDMM_taskFxn(UArg a0, UArg a1)
{
	uint8_t initialAdvertEnable = TRUE;
	BLE_init();

  // Application main loop
  for (;;)
  {
    // Waits for a signal to the semaphore associated with the calling thread.
    // Note that the semaphore associated with a thread is signaled when a
    // message is queued to the message receive queue of the thread or when
    // ICall_signal() function is called onto the semaphore.
    ICall_Errno errno = ICall_wait(ICALL_TIMEOUT_FOREVER);

    if (errno == ICALL_ERRNO_SUCCESS)
    {
      ICall_EntityID dest;
      ICall_ServiceEnum src;
      ICall_HciExtEvt *pMsg = NULL;

      // Check if we got a signal because of a stack message
      if (ICall_fetchServiceMsg(&src, &dest,
                                (void **)&pMsg) == ICALL_ERRNO_SUCCESS)
      {
        uint8 safeToDealloc = TRUE;

        if ((src == ICALL_SERVICE_CLASS_BLE) && (dest == selfEntity))
        {
          ICall_Event *pEvt = (ICall_Event *)pMsg;

          // Check for event flags received (event signature 0xffff)
          if (pEvt->signature == 0xffff)
          {
            // Event received when a connection event is completed
            if (pEvt->event_flag & WDMM_CONN_EVT_END_EVT)
            {
              // Try to retransmit pending ATT Response (if any)
              WDMM_sendAttRsp();
            }
          }
          else // It's a message from the stack and not an event.
          {
            // Process inter-task message
            safeToDealloc = WDMM_processStackMsg((ICall_Hdr *)pMsg);
          }
        }

        if (pMsg && safeToDealloc)
        {
          ICall_freeMsg(pMsg);
        }
      }

      // Process messages sent from another task or another context.
      while (!Queue_empty(hApplicationMsgQ))
      {
        app_msg_t *pMsg = Queue_dequeue(hApplicationMsgQ);

        // Process application-layer message probably sent from ourselves.
        user_processApplicationMessage(pMsg);

        // Free the received message.
        ICall_free(pMsg);
      }
    }

    if (events & ADC_READ_READY_EVT)
    {
        	events &= ~ADC_READ_READY_EVT;

        	WdmmService_SetParameter(WS_DC_READING_ID, WS_DC_READING_LEN, &send_dc_value);
        	WdmmService_SetParameter(WS_AC_READING_ID, WS_AC_READING_LEN, &send_ac_value);
    }

    if(events != 0)
    {
    		if (events & WDMM_BATTERY_REPORT_EVT)
    		{
    			events &= ~WDMM_BATTERY_REPORT_EVT;

    			WdmmService_SetParameter(WS_BATTERY_VOLTAGE_LEVEL_ID, WS_BATTERY_VOLTAGE_LEVEL_LEN, &battery_voltage);
    			WdmmService_SetParameter(WS_BATTERY_CAPACITY_LEVEL_ID, WS_BATTERY_CAPACITY_LEVEL_LEN, &state_of_charge);

    			Bq27426FuelGaugeService_SetParameter(BFGS_FLAGS_ID, BFGS_FLAGS_LEN, &bq27426_flags);
    			Bq27426FuelGaugeService_SetParameter(BFGS_AVERAGE_CURRENT_ID, BFGS_AVERAGE_CURRENT_LEN, &average_current);
    			Bq27426FuelGaugeService_SetParameter(BFGS_VOLTAGE_ID, BFGS_VOLTAGE_LEN, &battery_voltage);
    			Bq27426FuelGaugeService_SetParameter(BFGS_NOMINAL_AVAILABLE_CAPACITY_ID, BFGS_NOMINAL_AVAILABLE_CAPACITY_LEN, &nominal_available_capacity);
    			Bq27426FuelGaugeService_SetParameter(BFGS_FULL_AVAILABLE_CAPACITY_ID, BFGS_FULL_AVAILABLE_CAPACITY_LEN, &full_available_capacity);
    			Bq27426FuelGaugeService_SetParameter(BFGS_REMAINING_CAPACITY_ID, BFGS_REMAINING_CAPACITY_LEN, &remaining_capacity);
    			Bq27426FuelGaugeService_SetParameter(BFGS_FULL_CHARGE_CAPACITY_ID, BFGS_FULL_CHARGE_CAPACITY_LEN, &full_charge_capacity);
    			Bq27426FuelGaugeService_SetParameter(BFGS_AVERAGE_POWER_ID, BFGS_AVERAGE_POWER_LEN, &average_power);
    			Bq27426FuelGaugeService_SetParameter(BFGS_STATE_OF_CHARGE_ID, BFGS_STATE_OF_CHARGE_LEN, &state_of_charge);
     			Bq27426FuelGaugeService_SetParameter(BFGS_STATE_OF_HEALTH_ID, BFGS_STATE_OF_HEALTH_LEN, &state_of_health);
     			Bq27426FuelGaugeService_SetParameter(BFGS_TEMPERATURE_ID, BFGS_TEMPERATURE_LEN, &temperature);
//     			Bq27426FuelGaugeService_SetParameter(BFGS_INTERNAL_TEMPERATURE_ID, BFGS_INTERNAL_TEMPERATURE_LEN, &internal_temperature);
//  			Bq27426FuelGaugeService_SetParameter(BFGS_REMAINING_CAPACITY_UNFILTERED_ID, BFGS_REMAINING_CAPACITY_UNFILTERED_LEN, &remaining_capacity_unfiltered);
//  			Bq27426FuelGaugeService_SetParameter(BFGS_REMAINING_CAPACITY_FILTERED_ID, BFGS_REMAINING_CAPACITY_FILTERED_LEN, &remaining_capacity_filtered);
//  			Bq27426FuelGaugeService_SetParameter(BFGS_FULL_CHARGE_CAPACITY_UNFILTERED_ID, BFGS_FULL_CHARGE_CAPACITY_UNFILTERED_LEN, &full_charge_capacity_unfiltered);
//  			Bq27426FuelGaugeService_SetParameter(BFGS_FULL_CHARGE_CAPACITY_FILTERED_ID, BFGS_FULL_CHARGE_CAPACITY_FILTERED_LEN, &full_charge_capacity_filtered);
//  			Bq27426FuelGaugeService_SetParameter(BFGS_STATE_OF_CHARGE_UNFILTERED_ID, BFGS_STATE_OF_CHARGE_UNFILTERED_LEN, &state_of_charge_unfiltered);

    		}
    		if (events & WDMM_CAPTIVATE_SENSING_EVT)
    		{
    			events &= ~WDMM_CAPTIVATE_SENSING_EVT;
    			WdmmService_SetParameter(WS_CAPTIVATE_SENSING_ID, WS_CAPTIVATE_SENSING_LEN, &captivate_active);
    		}

    		if (events & WDMM_POWER_DOWN_EVT)
    		{
    			events &= ~WDMM_POWER_DOWN_EVT;
    			initialAdvertEnable = FALSE;
    			GAPRole_SetParameter(GAPROLE_ADVERT_ENABLED, sizeof(uint8_t),
       	                       &initialAdvertEnable);
    		}

    		if (events & WDMM_POWER_UP_EVT)
    		{
    			events &= ~WDMM_POWER_UP_EVT;
    			initialAdvertEnable = TRUE;
    			GAPRole_SetParameter(GAPROLE_ADVERT_ENABLED, sizeof(uint8_t),
           	                   &initialAdvertEnable);

    		}
    		if (events & WDMM_NFC_EVT)
    		{
    			events &= ~WDMM_NFC_EVT;
				RF430_configureOOBPairing();
				RF430_enableRF();
    		}
    }
  }
}

static void WDMM_adcreadTaskFxn(UArg a0, UArg a1)
{
		volatile uint8_t buffer=0;
		volatile uint16_t sample_count = 0;
		volatile int64_t sample;
		volatile int64_t dc_sum, ac_sum;
		volatile int16_t i=1;

		// SPI variables
		uint8_t  txBuffer[384];
		uint8_t  rxBuffer[2][384];

		SPI_Transaction spiTransaction;

		spiTransaction.count = 384;
		spiTransaction.txBuf = txBuffer;
		spiTransaction.rxBuf = rxBuffer[0];

		for(i=0;i<384;i+=3)							// set-up spi transmit buffer for ADS8885 CNVT
		{
			txBuffer[i]=0xFC;
			txBuffer[i+1]=0x00;
			txBuffer[i+2]=0x00;
		}

		SPI_transfer(spi_adc, &spiTransaction);

		PIN_setOutputValue(power_controlPinHandle, AFE_DCDC_ENABLE,1);
		PIN_setOutputValue(power_controlPinHandle, AFE_LDO_ENABLE,1);
		PIN_setOutputValue(afe_controlPinHandle, ADS8885_CNVST,1);

		while(1)
		{
			dc_sum=0;
			ac_sum=0;
			for(sample_count=0;sample_count<256;sample_count++)
			{

				Semaphore_pend(Semaphore_handle(&spiTaskAlert), BIOS_WAIT_FOREVER);

				spiTransaction.rxBuf = rxBuffer[buffer];

				SPI_transfer(spi_adc, &spiTransaction);

				buffer = !buffer;
				if(wdmm_connected == TRUE)
				{
					for(i=0;i<384;i=i+3)
					{
						if(((rxBuffer[buffer][i]) & 0x02)==0)
							sample = ((rxBuffer[buffer][i] & 0x03) << 16);
						else
							sample = ((rxBuffer[buffer][i] | 0xFFFC) << 16);

						sample = sample | (rxBuffer[buffer][i+1] << 8);
						sample = sample | (rxBuffer[buffer][i+2]);
						sample = -sample; 			// need to invert polarity of the ADC inputs

						dc_sum = dc_sum + sample;
						ac_sum = (sample * sample) + ac_sum;
					}
				}
			}

			if(wdmm_connected == TRUE)
			{
				dc_sum = (dc_sum >> 15);
				send_dc_value = dc_sum;
				ac_sum = (ac_sum >> 15);
				send_ac_value = ac_sum;

				events |= ADC_READ_READY_EVT;

				// Wake up the application.
				Semaphore_post(sem);
			}
	}
}

static void WDMM_pwr_mgmtTaskFxn(UArg a0, UArg a1)
{
		volatile uint16_t i;
		Bool previous_wdmm_connected=TRUE;
		Util_startClock(&periodicClock);

// ******************************************************************
// I2C Initialization
// ******************************************************************
		I2C_init();
		I2C_Params      i2cParams;
		I2C_Params_init(&i2cParams);
		i2c = I2C_open(WDMM_I2C1, &i2cParams);         //open the BQ27426
		if(i2c == NULL) {
			//          System_abort("Error Initializing i2c\n");
			while(i){}
		}

		bq27426_flags = BQ27426_cmd_read(0x06);
		if((bq27426_flags & 0x0020) > 0)				// re-initialize fuel gauge if POR detected
		{
			BQ27426_reset();
			BQ27426_configure();
		}

		while(1)
		{
			Semaphore_pend(Semaphore_handle(&pwr_mgmt_TaskAlert), BIOS_WAIT_FOREVER);
			PIN_setConfig(power_controlPinHandle, PIN_BM_IRQ, CC2640_WAKE_UP | PIN_IRQ_POSEDGE);

			battery_state_update_count++;

			if(wdmm_connected==TRUE)
			{
				apd_count = 0;
				PIN_setConfig(rf430_controlPinHandle, PIN_BM_IRQ, RF430_VCORE | PIN_IRQ_DIS);					//disable NFC interrupt when connected
				if(previous_wdmm_connected==FALSE)
				{
						previous_wdmm_connected = TRUE;
						battery_state_update_count = BATTERY_CHECK_INTERVAL_COUNT-2;

						I2C_close(i2c);
						i2c = I2C_open(WDMM_I2C1, &i2cParams);		// open the BQ27426
						if(i2c == NULL) {
							//          System_abort("Error Initializing i2c\n");
							while(1){}
						}
				}
			}
			else
			{
				if(expedite_power_down==TRUE)
				{
					apd_count=APD_TIME_INTERVAL_COUNT-5;
					expedite_power_down=FALSE;
				}
				apd_count++;
				i=apd_count;
				PIN_setConfig(rf430_controlPinHandle, PIN_BM_IRQ, RF430_VCORE | PIN_IRQ_POSEDGE);					//enable interrupt when not connected
				if(previous_wdmm_connected==TRUE)
				{
					previous_wdmm_connected = FALSE;

					I2C_close(i2c);
					i2c = I2C_open(WDMM_I2C0, &i2cParams);   // open RF430
					if(i2c == NULL) {
						//          System_abort("Error Initializing i2c\n");
						while(1){}
					}
					RF430_hardReset();
				}
			}

			if((previous_wdmm_connected==FALSE) && (wdmm_connected==FALSE) && (nfc_detected==TRUE))
			{
					nfc_detected=FALSE;
					events |= WDMM_NFC_EVT;
					Semaphore_post(sem);
			}

			if(apd_count>=APD_TIME_INTERVAL_COUNT)
			{
				wdmm_power_down();
			}

			if(battery_state_update_count>=BATTERY_CHECK_INTERVAL_COUNT)
			{
					battery_state_update_count=0;
					if(wdmm_connected==TRUE)
					{
						temperature = BQ27426_cmd_read(0x02);
						battery_voltage = BQ27426_cmd_read(0x04);
						nominal_available_capacity = BQ27426_cmd_read(0x08);
						full_available_capacity = BQ27426_cmd_read(0x0A);
						remaining_capacity = BQ27426_cmd_read(0x0C);
						full_charge_capacity = BQ27426_cmd_read(0x0E);
						average_current = BQ27426_cmd_read(0x10);
						average_power = BQ27426_cmd_read(0x18);
						state_of_charge = BQ27426_cmd_read(0x1C);
						bq27426_flags = BQ27426_cmd_read(0x06);
						state_of_health = BQ27426_cmd_read(0x20);
//						internal_temperature = BQ27426_cmd_read(0x1E);
//						remaining_capacity_unfiltered = BQ27426_cmd_read(0x28);
//						remaining_capacity_filtered = BQ27426_cmd_read(0x2A);
//						full_charge_capacity_unfiltered = BQ27426_cmd_read(0x2C);
//						full_charge_capacity_filtered = BQ27426_cmd_read(0x2E);
//						state_of_charge_unfiltered = BQ27426_cmd_read(0x30);

						events |= WDMM_BATTERY_REPORT_EVT;
						Semaphore_post(sem);

						if((bq27426_flags & 0x0020) > 0)
						{
					    	  BQ27426_reset();
					    	  BQ27426_configure();
						}
					}
			}
		}
}

void wdmm_power_down(void)
{
		volatile uint32_t k;

		apd_count=0;
		events |= WDMM_POWER_DOWN_EVT;
		Semaphore_post(sem);

		Util_stopClock(&periodicClock);

		Semaphore_pend(Semaphore_handle(&spiTaskAlert), BIOS_WAIT_FOREVER);
		for(k=0;k<0x3FFFFF;k++)											// allow time for all tasks to complete
		{
			;
		}

		PIN_setOutputValue(power_controlPinHandle, AFE_DCDC_ENABLE,0);
		PIN_setOutputValue(power_controlPinHandle, AFE_LDO_ENABLE,0);
		PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_1,0);
		PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_0,0);
		PIN_setOutputValue(afe_controlPinHandle, CURRENT_RANGE_SELECT,0);
		PIN_setOutputValue(afe_controlPinHandle, CURRENT_SHUNT_ISOLATE,0);
		PIN_setConfig(switchPinHandle, PIN_BM_IRQ, VOLTAGE_RANGE_SWITCH | PIN_IRQ_DIS);
		PIN_setConfig(switchPinHandle, PIN_BM_IRQ, MEASUREMENT_MODE_SWITCH | PIN_IRQ_DIS);

		Semaphore_pend(Semaphore_handle(&pwr_mgmt_TaskAlert), BIOS_WAIT_FOREVER);

		PIN_setOutputValue(power_controlPinHandle, AFE_DCDC_ENABLE,1);
		PIN_setOutputValue(power_controlPinHandle, AFE_LDO_ENABLE,1);

		PIN_setConfig(switchPinHandle, PIN_BM_IRQ, VOLTAGE_RANGE_SWITCH | PIN_IRQ_BOTHEDGES);
		PIN_setConfig(switchPinHandle, PIN_BM_IRQ, MEASUREMENT_MODE_SWITCH | PIN_IRQ_BOTHEDGES);
		Util_startClock(&periodicClock);
		Semaphore_post(Semaphore_handle(&spiTaskAlert));
		events |= WDMM_POWER_UP_EVT;
		Semaphore_post(sem);
}

void WDMM_HW_init(void)
{
	int i=1;
	uint8_t state;

	// ******************************************************************
	  // Hardware initialization
	  // ******************************************************************

	// Create one-shot clocks for internal periodic events
	  Util_constructClock(&periodicClock, wdmmPeriodicTimerAlertCallback,
	                          0, WDMM_PERIODIC_EVT_PERIOD, true, WDMM_PERIODIC_EVT);


	  // ******************************************************************
	  // SPI Initialization
	  // ******************************************************************

	  SPI_Params      spiParams;
	  SPI_init();
	  SPI_Params_init(&spiParams);
	  spiParams.dataSize = 8;
	  spiParams.bitRate = 5800000;
	  spiParams.transferMode = SPI_MODE_CALLBACK;
	  spiParams.transferCallbackFxn = transferCallback;
	  spiParams.frameFormat = SPI_POL1_PHA0;
	  spi_adc = SPI_open(WDMM_SPI0, &spiParams);
	  if (spi_adc == NULL) {
	  //          System_abort("Error Initializing SPI\n");
	     		while(i){}
	  }

// Create GPIO pin handles
	  rf430_controlPinHandle = PIN_open(&rf430_controlPinState, rf430_controlPinTable);
	  if(!rf430_controlPinHandle) {
	      Task_exit();
	  }

	  afe_controlPinHandle = PIN_open(&afe_controlPinState, afe_controlPinTable);
	  if(!afe_controlPinHandle) {
	    Task_exit();
	  }

	  switchPinHandle = PIN_open(&switchPinState, switchPinTable);
	  if(!switchPinHandle) {
	    Task_exit();
	  }

	  power_controlPinHandle = PIN_open(&power_controlPinState, power_controlPinTable);
	  if(!power_controlPinHandle) {
	    Task_exit();
	  }

// Setup callback for switch pins
	  if (PIN_registerIntCb(switchPinHandle, &switchCallbackFxn) != 0) {
	    Task_exit();
	  }

// Setup callback for power-up pins
	      if (PIN_registerIntCb(power_controlPinHandle, &powerUpCallbackFxn) != 0) {
	        Task_exit();
	      }
// Setup callback for power-up pins
   	      if (PIN_registerIntCb(rf430_controlPinHandle, &rf430CallbackFxn) != 0) {
   	        Task_exit();
   	      }

	 // Bring Captivate MSP430 and RF430 out of reset
      PIN_setOutputValue(power_controlPinHandle, MSP430_RESET,1);
      PIN_setOutputValue(rf430_controlPinHandle, RF430_nRST, 1);

      PIN_setOutputValue(power_controlPinHandle, ENABLE_BQ27426,1);

// Initialize I-Sel pin to the proper state
      state=!PIN_getInputValue(MEASUREMENT_MODE_SWITCH);   						// 1-voltage, 0-current
      if(state==0)
       	  PIN_setOutputValue(afe_controlPinHandle, CURRENT_SHUNT_ISOLATE,1);
      else
      	  PIN_setOutputValue(afe_controlPinHandle, CURRENT_SHUNT_ISOLATE,0);

	  // Create the debounce clock objects for switch 0 and switch 1
	  Clock_Params clockParams;
	  Clock_Params_init(&clockParams);

	  // Both clock objects use the same callback, so differentiate on argument
	  // given to the callback in Swi context
	  clockParams.arg = VOLTAGE_RANGE_SWITCH;

	  // Initialize to 50 ms timeout when Clock_start is called.
	  // Timeout argument is in ticks, so convert from ms to ticks via tickPeriod.
	  Clock_construct(&switch0DebounceClock, switchDebounceSwiFxn,
	                  50 * (1000/Clock_tickPeriod),
	                  &clockParams);

	  // Second switch
	  clockParams.arg = MEASUREMENT_MODE_SWITCH;
	  Clock_construct(&switch1DebounceClock, switchDebounceSwiFxn,
	                  50 * (1000/Clock_tickPeriod),
	                  &clockParams);

}
/*
 * @brief   Called before the task loop and contains application-specific
 *          initialization of the BLE stack, hardware setup, power-state
 *          notification if used, and BLE profile/service initialization.
 *
 * @param   None.
 *
 * @return  None.
 */
void BLE_init(void)
{
	uint8_t buf[12];
	volatile uint8_t status = SUCCESS;

  // ******************************************************************
  // NO STACK API CALLS CAN OCCUR BEFORE THIS CALL TO ICall_registerApp
  // ******************************************************************
  // Register the current thread as an ICall dispatcher application
  // so that the application can send and receive messages via ICall to Stack.
  ICall_registerApp(&selfEntity, &sem);

  // Initialize queue for application messages.
  // Note: Used to transfer control to application thread from e.g. interrupts.
  Queue_construct(&applicationMsgQ, NULL);
  hApplicationMsgQ = Queue_handle(&applicationMsgQ);

  // ******************************************************************
  // BLE Stack initialization
  // ******************************************************************

  // Setup the GAP Peripheral Role Profile
  uint8_t initialAdvertEnable = TRUE;  // Advertise on power-up

  // By setting this to zero, the device will go into the waiting state after
  // being discoverable. Otherwise wait this long [ms] before advertising again.
  uint16_t advertOffTime = 0; // miliseconds

  // Set advertisement enabled.
  GAPRole_SetParameter(GAPROLE_ADVERT_ENABLED, sizeof(uint8_t),
                       &initialAdvertEnable);

  // Configure the wait-time before restarting advertisement automatically
  GAPRole_SetParameter(GAPROLE_ADVERT_OFF_TIME, sizeof(uint16_t),
                       &advertOffTime);

  // Initialize Scan Response data
  GAPRole_SetParameter(GAPROLE_SCAN_RSP_DATA, sizeof(scanRspData), scanRspData);

  // Initialize Advertisement data
  GAPRole_SetParameter(GAPROLE_ADVERT_DATA, sizeof(advertData), advertData);

  // Set advertising interval
  uint16_t advInt = DEFAULT_ADVERTISING_INTERVAL;

  GAP_SetParamValue(TGAP_LIM_DISC_ADV_INT_MIN, advInt);
  GAP_SetParamValue(TGAP_LIM_DISC_ADV_INT_MAX, advInt);
  GAP_SetParamValue(TGAP_GEN_DISC_ADV_INT_MIN, advInt);
  GAP_SetParamValue(TGAP_GEN_DISC_ADV_INT_MAX, advInt);

  // Set duration of advertisement before stopping in Limited adv mode.
  GAP_SetParamValue(TGAP_LIM_ADV_TIMEOUT, 30); // Seconds

  // ******************************************************************
  // BLE Bond Manager initialization
  // ******************************************************************
  uint32_t passkey = 0; // passkey "000000"
  uint8_t pairMode = GAPBOND_PAIRING_MODE_WAIT_FOR_REQ;
  uint8_t mitm = TRUE;
  uint8_t ioCap = GAPBOND_IO_CAP_DISPLAY_ONLY;
  uint8_t bonding = TRUE;

  GAPBondMgr_SetParameter(GAPBOND_DEFAULT_PASSCODE, sizeof(uint32_t),
                          &passkey);
  GAPBondMgr_SetParameter(GAPBOND_PAIRING_MODE, sizeof(uint8_t), &pairMode);
  GAPBondMgr_SetParameter(GAPBOND_MITM_PROTECTION, sizeof(uint8_t), &mitm);
  GAPBondMgr_SetParameter(GAPBOND_IO_CAPABILITIES, sizeof(uint8_t), &ioCap);
  GAPBondMgr_SetParameter(GAPBOND_BONDING_ENABLED, sizeof(uint8_t), &bonding);

  // ******************************************************************
  // BLE Service initialization
  // ******************************************************************

  // Add services to GATT server
  GGS_AddService(GATT_ALL_SERVICES);           // GAP
  GATTServApp_AddService(GATT_ALL_SERVICES);   // GATT attributes
  DevInfo_AddService();                        // Device Information Service

  // Set the device name characteristic in the GAP Profile
  GGS_SetParameter(GGS_DEVICE_NAME_ATT, GAP_DEVICE_NAME_LEN, attDeviceName);

  // Add services to GATT server and give ID of this task for Indication acks.
  WdmmService_AddService();
  Bq27426FuelGaugeService_AddService();

  // Register callbacks with the generated services that
  // can generate events (writes received) to the application
  WdmmService_RegisterAppCBs( &user_Wdmm_ServiceCBs );
  Bq27426FuelGaugeService_RegisterAppCBs( &user_Bq27426_ServiceCBs );

  // Placeholder variable for characteristic intialization

  // Initalization of characteristics in WDMM_Service that can provide data.
  status = osal_snv_read(SNV_ID_VOLTAGE_50mV, sizeof(buf), buf);
  WdmmService_SetParameter(WS_VOLTAGE_RANGE_1_CAL_CONSTANTS_ID, WS_VOLTAGE_RANGE_1_CAL_CONSTANTS_LEN, buf);

  status = osal_snv_read(SNV_ID_VOLTAGE_500mV, sizeof(buf), buf);
  WdmmService_SetParameter(WS_VOLTAGE_RANGE_2_CAL_CONSTANTS_ID, WS_VOLTAGE_RANGE_2_CAL_CONSTANTS_LEN, buf);

  status = osal_snv_read(SNV_ID_VOLTAGE_5V, sizeof(buf), buf);
  WdmmService_SetParameter(WS_VOLTAGE_RANGE_3_CAL_CONSTANTS_ID, WS_VOLTAGE_RANGE_3_CAL_CONSTANTS_LEN, buf);

  status = osal_snv_read(SNV_ID_VOLTAGE_50V, sizeof(buf), buf);
  WdmmService_SetParameter(WS_VOLTAGE_RANGE_4_CAL_CONSTANTS_ID, WS_VOLTAGE_RANGE_4_CAL_CONSTANTS_LEN, buf);

  status = osal_snv_read(SNV_ID_CURRENT_500uA, sizeof(buf), buf);
  WdmmService_SetParameter(WS_CURRENT_RANGE_1_CAL_CONSTANTS_ID, WS_VOLTAGE_RANGE_1_CAL_CONSTANTS_LEN, buf);

  status = osal_snv_read(SNV_ID_CURRENT_50mA, sizeof(buf), buf);
  WdmmService_SetParameter(WS_CURRENT_RANGE_2_CAL_CONSTANTS_ID, WS_VOLTAGE_RANGE_2_CAL_CONSTANTS_LEN, buf);

  switch0State=!PIN_getInputValue(VOLTAGE_RANGE_SWITCH);
  WdmmService_SetParameter(WS_VOLTAGE_RANGE_MODE_ID, WS_VOLTAGE_RANGE_MODE_LEN, &switch0State);

  switch1State=!PIN_getInputValue(MEASUREMENT_MODE_SWITCH);   						// 1-voltage, 0-current
  WdmmService_SetParameter(WS_MEASUREMENT_MODE_ID, WS_MEASUREMENT_MODE_LEN, &switch1State);
  if(switch1State==0)
	  PIN_setOutputValue(afe_controlPinHandle, CURRENT_SHUNT_ISOLATE,1);
  else
	  PIN_setOutputValue(afe_controlPinHandle, CURRENT_SHUNT_ISOLATE,0);

  // Start the stack in Peripheral mode.
  VOID GAPRole_StartDevice(&user_gapRoleCBs);

  // Start Bond Manager
  VOID GAPBondMgr_Register(&user_bondMgrCBs);

  // Register with GAP for HCI/Host messages
  GAP_RegisterForMsgs(selfEntity);

  // Register for GATT local events and ATT Responses pending for transmission
  GATT_RegisterForMsgs(selfEntity);
}

/*
 * @brief   Handle application messages
 *
 *          These are messages not from the BLE stack, but from the
 *          application itself.
 *
 *          For example, in a Software Interrupt (Swi) it is not possible to
 *          call any BLE APIs, so instead the Swi function must send a message
 *          to the application Task for processing in Task context.
 *
 * @param   pMsg  Pointer to the message of type app_msg_t.
 *
 * @return  None.
 */
static void user_processApplicationMessage(app_msg_t *pMsg)
{
  char_data_t *pCharData = (char_data_t *)pMsg->pdu;

  switch (pMsg->type)
  {
    case APP_MSG_SERVICE_WRITE: /* Message about received value write */
      /* Call different handler per service */
      switch(pCharData->svcUUID) {
        case WDMM_SERVICE_SERV_UUID:
          user_WdmmService_ValueChangeHandler(pCharData);
          break;
        case BQ27426_FUEL_GAUGE_SERVICE_SERV_UUID:
          user_Bq27426FuelGaugeService_ValueChangeHandler(pCharData);
          break;
      }
      break;

    case APP_MSG_SERVICE_CFG: /* Message about received CCCD write */
      /* Call different handler per service */
      switch(pCharData->svcUUID) {
         }
      break;

    case APP_MSG_UPDATE_CHARVAL: /* Message from ourselves to send  */
//      user_updateCharVal(pCharData);
      break;

    case APP_MSG_GAP_STATE_CHANGE: /* Message that GAP state changed  */
      user_processGapStateChangeEvt( *(gaprole_States_t *)pMsg->pdu );
      break;

    case APP_MSG_SEND_PASSCODE: /* Message about pairing PIN request */
      {
        passcode_req_t *pReq = (passcode_req_t *)pMsg->pdu;
        // Send passcode response.
        GAPBondMgr_PasscodeRsp(pReq->connHandle, SUCCESS, DEFAULT_PASSCODE);
      }
      break;

    case APP_MSG_SWITCH_DEBOUNCED: /* Message from swi about pin change */
      {
        switch_state_t *pSwitchState = (switch_state_t *)pMsg->pdu;
        user_handleSwitchChange(pSwitchState);
      }
      break;
  }
}

/******************************************************************************
 *****************************************************************************
 *
 *  Handlers of system/application events deferred to the user Task context.
 *  Invoked from the application Task function above.
 *
 *  Further down you can find the callback handler section containing the
 *  functions that defer their actions via messages to the application task.
 *
 ****************************************************************************
 *****************************************************************************/

/*
 * @brief   Process a pending GAP Role state change event.
 *
 * @param   newState - new state
 *
 * @return  None.
 */
static void user_processGapStateChangeEvt(gaprole_States_t newState)
{
  switch ( newState )
  {
    case GAPROLE_STARTED:
      {
        uint8_t ownAddress[B_ADDR_LEN];
        uint8_t systemId[DEVINFO_SYSTEM_ID_LEN];

        GAPRole_GetParameter(GAPROLE_BD_ADDR, ownAddress);

        // use 6 bytes of device address for 8 bytes of system ID value
        systemId[0] = ownAddress[0];
        systemId[1] = ownAddress[1];
        systemId[2] = ownAddress[2];

        // set middle bytes to zero
        systemId[4] = 0x00;
        systemId[3] = 0x00;

        // shift three bytes up
        systemId[7] = ownAddress[5];
        systemId[6] = ownAddress[4];
        systemId[5] = ownAddress[3];

        DevInfo_SetParameter(DEVINFO_SYSTEM_ID, DEVINFO_SYSTEM_ID_LEN, systemId);

        // Display device address
        char *cstr_ownAddress = Util_convertBdAddr2Str(ownAddress);
      }
      break;

    case GAPROLE_ADVERTISING:
      wdmm_connected = FALSE;
      break;

    case GAPROLE_CONNECTED:
      {
        uint8_t peerAddress[B_ADDR_LEN];

        GAPRole_GetParameter(GAPROLE_CONN_BD_ADDR, peerAddress);
        wdmm_connected = TRUE;

        char *cstr_peerAddress = Util_convertBdAddr2Str(peerAddress);
       }
      break;

    case GAPROLE_CONNECTED_ADV:
      wdmm_connected = TRUE;
      break;

    case GAPROLE_WAITING:
      wdmm_connected = FALSE;
      break;

    case GAPROLE_WAITING_AFTER_TIMEOUT:
      wdmm_connected = FALSE;
      break;

    case GAPROLE_ERROR:
      wdmm_connected = FALSE;
      break;

    default:
      break;
  }
}

/*
 * @brief   Handle a debounced switch or release in Task context.
 *          Invoked by the taskFxn based on a message received from a callback.
 *
 * @see     switchDebounceSwiFxn
 * @see     switchCallbackFxn
 *
 * @param   pState  pointer to switch_state_t message sent from debounce Swi.
 *
 * @return  None.
 */
static void user_handleSwitchChange(switch_state_t *pState)
{
  // Update the service with the new value.
  // Will automatically send notification/indication if enabled.
  switch (pState->pinId)
  {
    case VOLTAGE_RANGE_SWITCH:

    	WdmmService_SetParameter(WS_VOLTAGE_RANGE_MODE_ID,
                                 sizeof(pState->state),
                                 &pState->state);
      break;
    case MEASUREMENT_MODE_SWITCH:

    	if(pState->state == 1)
    	{
    	   		PIN_setOutputValue(afe_controlPinHandle, CURRENT_SHUNT_ISOLATE,1);  //state=1: current
       	}
    	else
    	{
        		PIN_setOutputValue(afe_controlPinHandle, CURRENT_SHUNT_ISOLATE,0);  //state=0: voltage
    	}

     	WdmmService_SetParameter(WS_MEASUREMENT_MODE_ID,
                                 sizeof(pState->state),
                                 &pState->state);
      break;
  }
}

void user_WdmmService_ValueChangeHandler(char_data_t *pCharData)
{
	volatile uint8_t status;
	uint8_t buf[12];
  // Value to hold the received string for printing via Log, as Log printouts
  // happen in the Idle task, and so need to refer to a global/static variable.

  switch (pCharData->paramID)
  {
     case WS_VOLTAGE_RANGE_SELECT_ID:
    	  SetVoltageRange(pCharData->data[0]);
    	  break;

     case WS_CURRENT_RANGE_SELECT_ID:
    	  SetCurrentRange(pCharData->data[0]);
    	  break;

     case WS_VOLTAGE_RANGE_1_CAL_CONSTANTS_ID:
    	  status = osal_snv_write(SNV_ID_VOLTAGE_50mV, pCharData->dataLen, pCharData->data);
    	  status = osal_snv_read(SNV_ID_VOLTAGE_50mV, 12, buf);
    	  status = osal_snv_read(SNV_ID_VOLTAGE_50mV, 12, buf);
    	  break;

     case WS_VOLTAGE_RANGE_2_CAL_CONSTANTS_ID:
    	  status = osal_snv_write(SNV_ID_VOLTAGE_500mV, pCharData->dataLen, pCharData->data);
    	  status = osal_snv_read(SNV_ID_VOLTAGE_500mV, 12, buf);
    	  status = osal_snv_read(SNV_ID_VOLTAGE_500mV, 12, buf);
    	  break;

     case WS_VOLTAGE_RANGE_3_CAL_CONSTANTS_ID:
    	  status = osal_snv_write(SNV_ID_VOLTAGE_5V, pCharData->dataLen, pCharData->data);
    	  status = osal_snv_read(SNV_ID_VOLTAGE_5V, 12, buf);
    	  status = osal_snv_read(SNV_ID_VOLTAGE_5V, 12, buf);
    	  break;

     case WS_VOLTAGE_RANGE_4_CAL_CONSTANTS_ID:
    	  status = osal_snv_write(SNV_ID_VOLTAGE_50V, pCharData->dataLen, pCharData->data);
    	  status = osal_snv_read(SNV_ID_VOLTAGE_50V, 12, buf);
    	  status = osal_snv_read(SNV_ID_VOLTAGE_50V, 12, buf);
    	  break;

     case WS_CURRENT_RANGE_1_CAL_CONSTANTS_ID:
    	  status = osal_snv_write(SNV_ID_CURRENT_500uA, pCharData->dataLen, pCharData->data);
    	  status = osal_snv_read(SNV_ID_CURRENT_500uA, 12, buf);
    	  status = osal_snv_read(SNV_ID_CURRENT_500uA, 12, buf);
    	  break;

     case WS_CURRENT_RANGE_2_CAL_CONSTANTS_ID:
    	  status = osal_snv_write(SNV_ID_CURRENT_50mA, pCharData->dataLen, pCharData->data);
    	  status = osal_snv_read(SNV_ID_CURRENT_50mA, 12, buf);
    	  status = osal_snv_read(SNV_ID_CURRENT_50mA, 12, buf);
    	  break;

     case WS_INITIATE_POWER_DOWN_ID:
    	   expedite_power_down=TRUE;
       	  break;

    default:
    	 return;
  }
}

void user_Bq27426FuelGaugeService_ValueChangeHandler(char_data_t *pCharData)
{
	volatile uint8_t status;
  // Value to hold the received string for printing via Log, as Log printouts
  // happen in the Idle task, and so need to refer to a global/static variable.

  switch (pCharData->paramID)
  {
     case BFGS_INITIALIZE_FUEL_GAUGE_ID:
    	  BQ27426_reset();
     	  break;

    default:
    	 return;
  }
}

void SetVoltageRange(char i)
{
  		switch (i)
		{
       	 	 case VoltageRange_50mV:
    	 		PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_1,0);
    	 		PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_0,0);
    	 		break;
    	  	 case VoltageRange_500mV:
    	 	  	PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_1,0);
    	 	  	PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_0,1);
    	 	  	break;
    	 	 case VoltageRange_5V:
    	 	    PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_1,1);
    	 	   	PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_0,0);
    	 	   	break;
    	 	 case VoltageRange_50V:
    	 	    PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_1,1);
    	 	   	PIN_setOutputValue(afe_controlPinHandle, VOLTAGE_RANGE_SELECT_0,1);
    	 	   	break;
    	 	 default:
    	 		 return;
    	 }
}

void SetCurrentRange(char i)
{
  		switch (i)
		{
  			case CurrentRange_500uA:
  				PIN_setOutputValue(afe_controlPinHandle, CURRENT_RANGE_SELECT,1);
  				break;
  			case CurrentRange_50mA:
  				PIN_setOutputValue(afe_controlPinHandle, CURRENT_RANGE_SELECT,0);
  			break;

  			default:
  				return;
    	 }
}

/*
 * @brief   Process an incoming BLE stack message.
 *
 *          This could be a GATT message from a peer device like acknowledgement
 *          of an Indication we sent, or it could be a response from the stack
 *          to an HCI message that the user application sent.
 *
 * @param   pMsg - message to process
 *
 * @return  TRUE if safe to deallocate incoming message, FALSE otherwise.
 */
static uint8_t WDMM_processStackMsg(ICall_Hdr *pMsg)
{
  uint8_t safeToDealloc = TRUE;

  switch (pMsg->event)
  {
    case GATT_MSG_EVENT:
      // Process GATT message
      safeToDealloc = WDMM_processGATTMsg((gattMsgEvent_t *)pMsg);
      break;

    case HCI_GAP_EVENT_EVENT:
      {
        // Process HCI message
        switch(pMsg->status)
        {
          case HCI_COMMAND_COMPLETE_EVENT_CODE:
            // Process HCI Command Complete Event
            break;

          default:
            break;
        }
      }
      break;

    default:
      // do nothing
      break;
  }

  return (safeToDealloc);
}

/*
 * @brief   Process GATT messages and events.
 *
 * @return  TRUE if safe to deallocate incoming message, FALSE otherwise.
 */
static uint8_t WDMM_processGATTMsg(gattMsgEvent_t *pMsg)
{
  // See if GATT server was unable to transmit an ATT response
  if (pMsg->hdr.status == blePending)
  {
    // No HCI buffer was available. Let's try to retransmit the response
    // on the next connection event.
    if (HCI_EXT_ConnEventNoticeCmd(pMsg->connHandle, selfEntity,
                                   WDMM_CONN_EVT_END_EVT) == SUCCESS)
    {
      // First free any pending response
      WDMM_freeAttRsp(FAILURE);

      // Hold on to the response message for retransmission
      pAttRsp = pMsg;

      // Don't free the response message yet
      return (FALSE);
    }
  }
  else if (pMsg->method == ATT_FLOW_CTRL_VIOLATED_EVENT)
  {
    // ATT request-response or indication-confirmation flow control is
    // violated. All subsequent ATT requests or indications will be dropped.
    // The app is informed in case it wants to drop the connection.


  }
  else if (pMsg->method == ATT_MTU_UPDATED_EVENT)
  {
    // MTU size updated

  }
  else
  {
    // Got an expected GATT message from a peer.

  }

  // Free message payload. Needed only for ATT Protocol messages
  GATT_bm_free(&pMsg->msg, pMsg->method);

  // It's safe to free the incoming message
  return (TRUE);
}

/*
 *  Application error handling functions
 *****************************************************************************/

/*
 * @brief   Send a pending ATT response message.
 *
 *          The message is one that the stack was trying to send based on a
 *          peer request, but the response couldn't be sent because the
 *          user application had filled the TX queue with other data.
 *
 * @param   none
 *
 * @return  none
 */
static void WDMM_sendAttRsp(void)
{
  // See if there's a pending ATT Response to be transmitted
  if (pAttRsp != NULL)
  {
    uint8_t status;

    // Increment retransmission count
    rspTxRetry++;

    // Try to retransmit ATT response till either we're successful or
    // the ATT Client times out (after 30s) and drops the connection.
    status = GATT_SendRsp(pAttRsp->connHandle, pAttRsp->method, &(pAttRsp->msg));
    if ((status != blePending) && (status != MSG_BUFFER_NOT_AVAIL))
    {
      // Disable connection event end notice
      HCI_EXT_ConnEventNoticeCmd(pAttRsp->connHandle, selfEntity, 0);

      // We're done with the response message
      WDMM_freeAttRsp(status);
    }
    else
    {
      // Continue retrying
     }
  }
}

/*
 * @brief   Free ATT response message.
 *
 * @param   status - response transmit status
 *
 * @return  none
 */
static void WDMM_freeAttRsp(uint8_t status)
{
  // See if there's a pending ATT response message
  if (pAttRsp != NULL)
  {
    // See if the response was sent out successfully
    if (status == SUCCESS)
    {

    }
    else
    {

      // Free response payload
      GATT_bm_free(&pAttRsp->msg, pAttRsp->method);
    }

    // Free response message
    ICall_freeMsg(pAttRsp);

    // Reset our globals
    pAttRsp = NULL;
    rspTxRetry = 0;
  }
}

/******************************************************************************
 *****************************************************************************
 *
 *  Handlers of direct system callbacks.
 *
 *  Typically enqueue the information or request as a message for the
 *  application Task for handling.
 *
 ****************************************************************************
 *****************************************************************************/


/*
 *  Callbacks from the Stack Task context (GAP or Service changes)
 *****************************************************************************/

/**
 * Callback from GAP Role indicating a role state change.
 */
static void user_gapStateChangeCB(gaprole_States_t newState)
{
  user_enqueueRawAppMsg( APP_MSG_GAP_STATE_CHANGE, (uint8_t *)&newState, sizeof(newState) );
}

/*
 * @brief   Passcode callback.
 *
 * @param   connHandle - connection handle
 * @param   uiInputs   - input passcode?
 * @param   uiOutputs  - display passcode?
 *
 * @return  none
 */
static void user_gapBondMgr_passcodeCB(uint8_t *deviceAddr, uint16_t connHandle,
                                       uint8_t uiInputs, uint8_t uiOutputs)
{
  passcode_req_t req =
  {
    .connHandle = connHandle,
    .uiInputs = uiInputs,
    .uiOutputs = uiOutputs
  };

  // Defer handling of the passcode request to the application, in case
  // user input is required, and because a BLE API must be used from Task.
  user_enqueueRawAppMsg(APP_MSG_SEND_PASSCODE, (uint8_t *)&req, sizeof(req));
}

/*
 * @brief   Pairing state callback.
 *
 * @param   connHandle - connection handle
 * @param   state      - pairing state
 * @param   status     - pairing status
 *
 * @return  none
 */
static void user_gapBondMgr_pairStateCB(uint16_t connHandle, uint8_t state,
                                        uint8_t status)
{
  if (state == GAPBOND_PAIRING_STATE_STARTED)
  {

  }
  else if (state == GAPBOND_PAIRING_STATE_COMPLETE)
  {
    if (status == SUCCESS)
    {

    }
    else
    {

    }
  }
  else if (state == GAPBOND_PAIRING_STATE_BONDED)
  {
    if (status == SUCCESS)
    {

    }
  }
}

/**
 * Callback handler for characteristic value changes in services.
 */
static void user_service_ValueChangeCB( uint16_t connHandle, uint16_t svcUuid,
                                        uint8_t paramID, uint8_t *pValue,
                                        uint16_t len )
{
  // See the service header file to compare paramID with characteristic.
  user_enqueueCharDataMsg(APP_MSG_SERVICE_WRITE, connHandle, svcUuid, paramID,
                          pValue, len);
}

/*
 *  Callbacks from Swi-context
 *****************************************************************************/

/*
 * @brief  Callback from Clock module on timeout
 *
 *         Determines new state after debouncing
 *
 * @param  switchId    The pin being debounced
 */
static void switchDebounceSwiFxn(UArg switchId)
{
  // Used to send message to app
  switch_state_t switchMsg = { .pinId = switchId };
  uint8_t        sendMsg   = FALSE;

  // Get current value of the switch pin after the clock timeout
  uint8_t switchPinVal = PIN_getInputValue(switchId);
  switchMsg.state=1;

  // Set interrupt direction to opposite of debounced state
  // If switch is now released (switch is active low, so release is high)
  if (switchPinVal)
  {
    // Enable interrupts
    PIN_setConfig(switchPinHandle, PIN_BM_IRQ, switchId | PIN_IRQ_BOTHEDGES);
  }
  else
  {
    // Enable interrupts
    PIN_setConfig(switchPinHandle, PIN_BM_IRQ, switchId | PIN_IRQ_BOTHEDGES);
  }

  switch(switchId)
  {
    case VOLTAGE_RANGE_SWITCH:
      if (switchPinVal && switch0State)
      {
        switchMsg.state = switch0State = 0;
        sendMsg = TRUE;
      }
      else if (!switchPinVal && !switch0State)
      {
        switchMsg.state = switch0State = 1;
        sendMsg = TRUE;
      }
      break;

    case MEASUREMENT_MODE_SWITCH:
      if (switchPinVal && switch1State)
      {
        switchMsg.state = switch1State = 0;
        sendMsg = TRUE;
      }
      else if (!switchPinVal && !switch1State)
      {
        switchMsg.state = switch1State = 1;
        sendMsg = TRUE;
      }
      break;
  }

  if (sendMsg == TRUE)
  {
    user_enqueueRawAppMsg(APP_MSG_SWITCH_DEBOUNCED,
                      (uint8_t *)&switchMsg, sizeof(switchMsg));
  }
}

/*
 *  Callbacks from Hwi-context
 *****************************************************************************/

/*
 * @brief  Callback from PIN driver on interrupt
 *
 *         Sets in motion the debouncing.
 *
 * @param  handle    The PIN_Handle instance this is about
 * @param  pinId     The pin that generated the interrupt
 */
static void switchCallbackFxn(PIN_Handle handle, PIN_Id pinId)
{
  // Disable interrupt on that pin for now. Re-enabled after debounce.
  PIN_setConfig(handle, PIN_BM_IRQ, pinId | PIN_IRQ_DIS);

  // Start debounce timer
  switch (pinId)
  {
    case VOLTAGE_RANGE_SWITCH:
      Clock_start(Clock_handle(&switch0DebounceClock));
      break;
    case MEASUREMENT_MODE_SWITCH:
      Clock_start(Clock_handle(&switch1DebounceClock));
      break;
  }
}

/******************************************************************************
 *****************************************************************************
 *
 *  Utility functions
 *
 ****************************************************************************
 *****************************************************************************/

/*
 * @brief  Generic message constructor for characteristic data.
 *
 *         Sends a message to the application for handling in Task context where
 *         the message payload is a char_data_t struct.
 *
 *         From service callbacks the appMsgType is APP_MSG_SERVICE_WRITE or
 *         APP_MSG_SERVICE_CFG, and functions running in another context than
 *         the Task itself, can set the type to APP_MSG_UPDATE_CHARVAL to
 *         make the user Task loop invoke user_updateCharVal function for them.
 *
 * @param  appMsgType    Enumerated type of message being sent.
 * @param  connHandle    GAP Connection handle of the relevant connection
 * @param  serviceUUID   16-bit part of the relevant service UUID
 * @param  paramID       Index of the characteristic in the service
 * @oaram  *pValue       Pointer to characteristic value
 * @param  len           Length of characteristic data
 */
static void user_enqueueCharDataMsg( app_msg_types_t appMsgType,
                                     uint16_t connHandle,
                                     uint16_t serviceUUID, uint8_t paramID,
                                     uint8_t *pValue, uint16_t len )
{
  // Called in Stack's Task context, so can't do processing here.
  // Send message to application message queue about received data.
  uint16_t readLen = len; // How much data was written to the attribute

  // Allocate memory for the message.
  // Note: The pCharData message doesn't have to contain the data itself, as
  //       that's stored in a variable in the service implementation.
  //
  //       However, to prevent data loss if a new value is received before the
  //       service's container is read out via the GetParameter API is called,
  //       we copy the characteristic's data now.
  app_msg_t *pMsg = ICall_malloc( sizeof(app_msg_t) + sizeof(char_data_t) +
                                  readLen );

  if (pMsg != NULL)
  {
    pMsg->type = appMsgType;

    char_data_t *pCharData = (char_data_t *)pMsg->pdu;
    pCharData->svcUUID = serviceUUID; // Use 16-bit part of UUID.
    pCharData->paramID = paramID;
    // Copy data from service now.
    memcpy(pCharData->data, pValue, readLen);
    // Update pCharData with how much data we received.
    pCharData->dataLen = readLen;
    // Enqueue the message using pointer to queue node element.
    Queue_enqueue(hApplicationMsgQ, &pMsg->_elem);
    // Let application know there's a message.
    Semaphore_post(sem);
  }
}

/*
 * @brief  Generic message constructor for application messages.
 *
 *         Sends a message to the application for handling in Task context.
 *
 * @param  appMsgType    Enumerated type of message being sent.
 * @oaram  *pValue       Pointer to characteristic value
 * @param  len           Length of characteristic data
 */
static void user_enqueueRawAppMsg(app_msg_types_t appMsgType, uint8_t *pData,
                                  uint16_t len)
{
  // Allocate memory for the message.
  app_msg_t *pMsg = ICall_malloc( sizeof(app_msg_t) + len );

  if (pMsg != NULL)
  {
    pMsg->type = appMsgType;

    // Copy data into message
    memcpy(pMsg->pdu, pData, len);

    // Enqueue the message using pointer to queue node element.
    Queue_enqueue(hApplicationMsgQ, &pMsg->_elem);
    // Let application know there's a message.
    Semaphore_post(sem);
  }
}

/*********************************************************************
*********************************************************************/
void RF430_hardReset(void)
{
	// RST RF430 (in case board remained powered but the MSP430 reset for some reason - MSP430 RST button pushed for example)
	PIN_setOutputValue(rf430_controlPinHandle, RF430_nRST, 0);		// Hold the reset line low
	delay_ms(2);
	PIN_setOutputValue(rf430_controlPinHandle, RF430_nRST, 1);		// Release the RF430CL330H from Reset
	delay_ms(20);
}

void RF430_enableRF(void)
{
	uint16_t ui16ControlReg;

	ui16ControlReg = RF430_readSingle(CONTROL_REG);

	ui16ControlReg |= RF_ENABLE;

	RF430_writeSingle(CONTROL_REG, ui16ControlReg);
}

/*******************************************************************************
 * @fn          RF430_configureOOBPairing
 *
 * @brief       Run a sensor self-test
 *
 * @return      true if passed
 ******************************************************************************/
void RF430_configureOOBPairing(void)
{
	uint8_t ui8Value;
	uint8_t pui8KeyBuffer[16];
	uint8_t ui8LoopCount;
    uint8_t ui8ownDeviceAddress[B_ADDR_LEN];

    GAPRole_GetParameter(GAPROLE_BD_ADDR, ui8ownDeviceAddress);

	// Set BLE Address value
	for (ui8LoopCount = 0; ui8LoopCount < 6; ui8LoopCount++)
	{
		NDEF_Application_Data[BLE_ADDRESS_START_INDEX+ui8LoopCount] = ui8ownDeviceAddress[ui8LoopCount];
	}

	// !!! Generate TK !!! - Current hardcoded for initial demo - RNG will be added shortly
	for (ui8LoopCount = 0; ui8LoopCount < 16; ui8LoopCount++)
	{
		pui8KeyBuffer[ui8LoopCount] = 0xAA;
	}

	// Pass TK to BLE Stack Bond Manager
	ui8Value = true;
	GAPBondMgr_SetParameter(GAPBOND_OOB_ENABLED, 1, &ui8Value);
	GAPBondMgr_SetParameter(GAPBOND_OOB_DATA, KEYLEN, pui8KeyBuffer);

	// Turn on MITM
	ui8Value = true;
	GAPBondMgr_SetParameter(GAPBOND_MITM_PROTECTION, 1, &ui8Value);

	// Enable No Input/No Output - this will force Just Works pairing if OOB pairing is not enabled by both devices
	ui8Value = GAPBOND_IO_CAP_NO_INPUT_NO_OUTPUT;
	GAPBondMgr_SetParameter(GAPBOND_IO_CAPABILITIES, 1, &ui8Value);

	// Set TK value in the NDEF message here
	for (ui8LoopCount = 0; ui8LoopCount < 16; ui8LoopCount++)
	{
		NDEF_Application_Data[TK_START_INDEX+ui8LoopCount] = pui8KeyBuffer[ui8LoopCount];
	}

	// Write RF430 NDEF Message with pairing information included
	RF430_writeContinuous(0, NDEF_Application_Data, 142);
}

/*******************************************************************************
 * @fn          RF430_writeSingle
 * @brief       This function implements the I2C protocol to write to a sensor.
 *              The sensor must be selected before this routine is called.
 *
 * @param       addr - which register to write
 * @param       pBuf - pointer to buffer containing data to be written
 * @param       nBytes - number of bytes to write
 *
 * @return      TRUE if successful write
 ******************************************************************************/

void RF430_writeSingle(uint16_t ui16Address, uint16_t ui16Value)
{
	I2C_Transaction i2cTransaction;
	uint8_t pui8TxBuffer[4];

	// Acquire I2C resource
	while (!Semaphore_pend(Semaphore_handle(&mutex),MS_2_TICKS(I2C_TIMEOUT)));

	pui8TxBuffer[0] = ui16Address >> 8; 		// MSB of address
	pui8TxBuffer[1] = ui16Address & 0xFF; 		// LSB of address
	pui8TxBuffer[2] = ui16Value & 0xFF; 		// LSB of data
	pui8TxBuffer[3] = ui16Value >> 8; 			// MSB of data

	i2cTransaction.writeBuf = pui8TxBuffer;
	i2cTransaction.writeCount = 4;
	i2cTransaction.readBuf = NULL;
	i2cTransaction.readCount = 0;
	i2cTransaction.slaveAddress = RF430_DEV_ADDR; //arbitrary for demo
	I2C_transfer(i2c, &i2cTransaction);

	// Release I2C resource
	Semaphore_post(Semaphore_handle(&mutex));
}


/*******************************************************************************
 * @fn          RF430_readSingle
 *
 * @brief       This function implements the I2C protocol to read from a sensor.
 *              The sensor must be selected before this routine is called.
 *
 * @param       addr - which register to read
 * @param       pBuf - pointer to buffer to place data
 * @param       nBytes - number of bytes to read
 *
 * @return      TRUE if the required number of bytes are received
 ******************************************************************************/
uint16_t RF430_readSingle(uint16_t ui16Address)
{
	uint16_t ui16Value;
	I2C_Transaction i2cTransaction;
	uint8_t pui8TxBuffer[2];
	uint8_t pui8RxBuffer[2];

	// Acquire I2C resource
	while (!Semaphore_pend(Semaphore_handle(&mutex),MS_2_TICKS(I2C_TIMEOUT)));

	pui8TxBuffer[0] = ui16Address >> 8; 			// MSB of address
	pui8TxBuffer[1] = ui16Address & 0xFF; 		// LSB of address

	i2cTransaction.writeBuf = pui8TxBuffer;
	i2cTransaction.writeCount = 2;
	i2cTransaction.readBuf = pui8RxBuffer;
	i2cTransaction.readCount = 2;
	i2cTransaction.slaveAddress = RF430_DEV_ADDR; //arbitrary for demo
	I2C_transfer(i2c, &i2cTransaction);

	// Release I2C resource
	Semaphore_post(Semaphore_handle(&mutex));

	ui16Value = pui8RxBuffer[1] << 8; 			// MSB of data
	ui16Value = pui8RxBuffer[0] && 0xFF; 		// LSB of data

	return ui16Value;
}

/*******************************************************************************
 * @fn          RF430_writeContinuous
 * @brief       This function implements the I2C protocol to write to a sensor.
 *              The sensor must be selected before this routine is called.
 *
 * @param       addr - which register to write
 * @param       pBuf - pointer to buffer containing data to be written
 * @param       nBytes - number of bytes to write
 *
 * @return      TRUE if successful write
 ******************************************************************************/

void RF430_writeContinuous(uint16_t ui16Address, uint8_t * ui8Data, uint16_t ui16Length)
{
	I2C_Transaction i2cTransaction;
	uint8_t pui8TxBuffer[2];

	// Acquire I2C resource
	while (!Semaphore_pend(Semaphore_handle(&mutex),MS_2_TICKS(I2C_TIMEOUT)));

	pui8TxBuffer[0] = ui16Address >> 8; 			// MSB of address
	pui8TxBuffer[1] = ui16Address & 0xFF; 		// LSB of address

	i2cTransaction.writeBuf = pui8TxBuffer;
	i2cTransaction.writeCount = 2;
	i2cTransaction.readBuf = NULL;
	i2cTransaction.readCount = 0;
	i2cTransaction.slaveAddress = RF430_DEV_ADDR; //arbitrary for demo
	I2C_transfer(i2c, &i2cTransaction);

	i2cTransaction.writeBuf = ui8Data;
	i2cTransaction.writeCount = ui16Length;
	i2cTransaction.readBuf = NULL;
	i2cTransaction.readCount = 0;
	i2cTransaction.slaveAddress = RF430_DEV_ADDR; //arbitrary for demo
	I2C_transfer(i2c, &i2cTransaction);

	// Release I2C resource
	Semaphore_post(Semaphore_handle(&mutex));
}

/*******************************************************************************
 * @fn          RF430_readContinuous
 * @brief       This function implements the I2C protocol to write to a sensor.
 *              The sensor must be selected before this routine is called.
 *
 * @param       addr - which register to write
 * @param       pBuf - pointer to buffer containing data to be written
 * @param       nBytes - number of bytes to write
 *
 * @return      TRUE if successful write
 ******************************************************************************/

void RF430_readContinuous(uint16_t ui16Address, uint8_t * ui8Data, uint16_t ui16Length)
{
	I2C_Transaction i2cTransaction;
	uint8_t pui8TxBuffer[2];

	// Acquire I2C resource
	while (!Semaphore_pend(Semaphore_handle(&mutex),MS_2_TICKS(I2C_TIMEOUT)));

	pui8TxBuffer[0] = ui16Address >> 8; 		// MSB of address
	pui8TxBuffer[1] = ui16Address & 0xFF; 		// LSB of address

	i2cTransaction.writeBuf = pui8TxBuffer;
	i2cTransaction.writeCount = 2;
	i2cTransaction.readBuf = NULL;
	i2cTransaction.readCount = 0;
	i2cTransaction.slaveAddress = RF430_DEV_ADDR; //arbitrary for demo
	I2C_transfer(i2c, &i2cTransaction);

	i2cTransaction.writeBuf = NULL;
	i2cTransaction.writeCount = 0;
	i2cTransaction.readBuf = ui8Data;
	i2cTransaction.readCount = ui16Length;
	i2cTransaction.slaveAddress = RF430_DEV_ADDR; //arbitrary for demo
	I2C_transfer(i2c, &i2cTransaction);

	// Release I2C resource
	Semaphore_post(Semaphore_handle(&mutex));
}

// nRegister = gauge register
// pData = pointer to data block which will hold the data from the gauge
// nLength = length
static BQ27426_read(uint8_t nRegister, uint8_t *pData, uint8_t nLength)
{
	I2C_Transaction i2cTransaction;
	uint8_t writeBuffer[1];
	volatile Bool i=1;
	volatile Bool transferOK;

	i2cTransaction.slaveAddress = BQ27426_DEVICE_ADDRESS; /* 7-bit peripheral slave address */
	i2cTransaction.writeBuf = writeBuffer; /* Buffer to be written */
	i2cTransaction.writeCount = 1; /* Number of bytes to be written */
	i2cTransaction.readBuf = NULL; /* Buffer to be read */
	i2cTransaction.readCount = 0; /* Number of bytes to be read */
	writeBuffer[0]=nRegister;

	transferOK = I2C_transfer(i2c, &i2cTransaction); /* Perform I2C transfer */
	if (!transferOK) {
					i = 0;
						/* I2C bus fault */
	}

	i2cTransaction.writeBuf = NULL; /* Buffer to be written */
	i2cTransaction.writeCount = 0; /* Number of bytes to be written */
	i2cTransaction.readBuf = pData; /* Buffer to be written */
	i2cTransaction.readCount = nLength; /* Number of bytes to be read */
	transferOK = I2C_transfer(i2c, &i2cTransaction); /* Perform I2C transfer */
	if (!transferOK) {
				i = 0;
					/* I2C bus fault */
	}
	delay_ms(10);
}

static BQ27426_write(uint8_t *pData, uint8_t nLength)
{
	I2C_Transaction i2cTransaction;
	volatile Bool i=1;
	volatile Bool transferOK;

	i2cTransaction.slaveAddress = BQ27426_DEVICE_ADDRESS; /* 7-bit peripheral slave address */
	i2cTransaction.writeBuf = pData; /* Buffer to be written */
	i2cTransaction.writeCount = nLength; /* Number of bytes to be written */
	i2cTransaction.readBuf = NULL; /* Buffer to be read */
	i2cTransaction.readCount = 0; /* Number of bytes to be read */

	transferOK = I2C_transfer(i2c, &i2cTransaction); /* Perform I2C transfer */
	if (!transferOK) {
					i = 0;
						/* I2C bus fault */
	}
	delay_ms(10);
}

// enter CFG_UPDATE mode
static void BQ27426_cfg_update_mode()
{
	uint16_t nFlags;
	uint8_t pData[3];

	pData[0]=0x00; pData[1]=0x13; pData[2]=0x00;
	BQ27426_write(pData,3);

	nFlags = BQ27426_cmd_read(0x06);
	while((nFlags & 0x0010)==0)
	{
		nFlags = BQ27426_cmd_read(0x06);
		delay_ms(100);
	}
	delay_ms(1200);
}

// exit CFG_UPDATE mode
static void BQ27426_soft_reset()
{
	uint16_t nFlags;
	uint8_t pData[3];

	pData[0]=0x00; pData[1]=0x42; pData[2]=0x00;
	BQ27426_write(pData,3);

	nFlags = BQ27426_cmd_read(0x06);
	while((nFlags & 0x0010)>0)
	{
		nFlags = BQ27426_cmd_read(0x06);
		delay_ms(100);
	}
}

// Full Reset
static void BQ27426_reset()
{
	uint8_t pData[3];

	pData[0]=0x00; pData[1]=0x41; pData[2]=0x00;
	BQ27426_write(pData,3);
	delay_ms(1000);
}

static uint16_t BQ27426_cmd_read(uint8_t nCmd)
{
	uint8_t pData[2];

	BQ27426_read(nCmd, pData, 2);
	delay_ms(10);

	return (pData[1] << 8) | pData[0];
}

static void BQ27426_configure()
{
	uint8_t pData[33];
	volatile uint8_t checksum=0, checksum2=0;

	BQ27426_cfg_update_mode();		// enter configuration update mode

	// Set Battery Profile to 3142
	pData[0]=0x00; pData[1]=0x32; pData[2]=0x00;
	BQ27426_write(pData,3);

	pData[0]=0x00; pData[1]=0x08; pData[2]=0x00;
	BQ27426_write(pData,3);
	BQ27426_read(0x00,pData,2);

	pData[0]=0x61; pData[1]=0x00;		// Enable access to configuration data
	BQ27426_write(pData,2);

//	Configure CURRENT Subclass - change deadband value to 1  ///
		pData[0]=0x3E; pData[1]=0x6B;
		BQ27426_write(pData,2);

		pData[0]=0x3F; pData[1]=0x00;
		BQ27426_write(pData,2);

		BQ27426_read(0x60,pData,1);
		checksum = pData[0];
		checksum = 255 - checksum;

		checksum = BQ27426_parameter_update(0x41, 0x01, checksum);
		checksum = 255 - checksum;

		pData[0]=0x60; pData[1]=checksum;
		BQ27426_write(pData,2);

//		BQ27426_read(0x40,pData,32);

#ifdef TR10440_HIGH_RESOLUTION
	BQ27426_subclass_update(Ra_Tables_SubClass_ID, Ra_Tables);
	BQ27426_subclass_update(Current_Thresholds_SubClass_ID, Current_Thresholds);
	BQ27426_subclass_update(State_SubClass_ID, State);
	BQ27426_subclass_update(IT_Cfg_0_SubClass_ID, IT_Cfg_0);
	BQ27426_subclass_update(IT_Cfg_1_SubClass_ID, IT_Cfg_1);
	BQ27426_subclass_update(IT_Cfg_2_SubClass_ID, IT_Cfg_2);
	BQ27426_subclass_update(Chem_Data_SubClass_ID, Chem_Data);
	BQ27426_subclass_update(Charge_Termination_SubClass_ID, Charge_Termination);
	BQ27426_subclass_update(Discharge_SubClass_ID, Discharge);
	BQ27426_subclass_update(Registers_SubClass_ID, Registers);
	BQ27426_subclass_update(Power_SubClass_ID, Power);
	BQ27426_subclass_update(OCVa_0_SubClass_ID, OCVa_0);
	BQ27426_subclass_update(OCVa_1_SubClass_ID, OCVa_1);
	BQ27426_subclass_update(OCVb_0_SubClass_ID, OCVb_0);
	BQ27426_subclass_update(OCVb_1_SubClass_ID, OCVb_1);
	BQ27426_subclass_update(Resistance_Low_Temp_SubClass_ID, Resistance_Low_Temp);
	BQ27426_subclass_update(Resistance_High_Temp_SubClass_ID, Resistance_High_Temp);
#endif

#ifdef TR10440_LOW_RESOLUTION
	BQ27426_subclass_update(State_SubClass_ID, State);
	BQ27426_subclass_update(Current_Thresholds_SubClass_ID, Current_Thresholds);
#endif

#ifdef LIR2477_HIGH_RESOLUTION
	BQ27426_subclass_update(Ra_Tables_SubClass_ID, Ra_Tables);
	BQ27426_subclass_update(Current_Thresholds_SubClass_ID, Current_Thresholds);
	BQ27426_subclass_update(State_SubClass_ID, State);
	BQ27426_subclass_update(IT_Cfg_0_SubClass_ID, IT_Cfg_0);
	BQ27426_subclass_update(IT_Cfg_1_SubClass_ID, IT_Cfg_1);
	BQ27426_subclass_update(IT_Cfg_2_SubClass_ID, IT_Cfg_2);
	BQ27426_subclass_update(Chem_Data_SubClass_ID, Chem_Data);
	BQ27426_subclass_update(Charge_Termination_SubClass_ID, Charge_Termination);
	BQ27426_subclass_update(Discharge_SubClass_ID, Discharge);
	BQ27426_subclass_update(Registers_SubClass_ID, Registers);
	BQ27426_subclass_update(Power_SubClass_ID, Power);
	BQ27426_subclass_update(OCVa_0_SubClass_ID, OCVa_0);
	BQ27426_subclass_update(OCVa_1_SubClass_ID, OCVa_1);
	BQ27426_subclass_update(OCVb_0_SubClass_ID, OCVb_0);
	BQ27426_subclass_update(OCVb_1_SubClass_ID, OCVb_1);
	BQ27426_subclass_update(Resistance_Low_Temp_SubClass_ID, Resistance_Low_Temp);
	BQ27426_subclass_update(Resistance_High_Temp_SubClass_ID, Resistance_High_Temp);
#endif

#ifdef LIR2477_LOW_RESOLUTION
	BQ27426_subclass_update(Ra_Tables_SubClass_ID, Ra_Tables);
	BQ27426_subclass_update(Current_Thresholds_SubClass_ID, Current_Thresholds);
	BQ27426_subclass_update(State_SubClass_ID, State);
	BQ27426_subclass_update(IT_Cfg_0_SubClass_ID, IT_Cfg_0);
	BQ27426_subclass_update(IT_Cfg_1_SubClass_ID, IT_Cfg_1);
	BQ27426_subclass_update(IT_Cfg_2_SubClass_ID, IT_Cfg_2);
	BQ27426_subclass_update(Chem_Data_SubClass_ID, Chem_Data);
	BQ27426_subclass_update(Charge_Termination_SubClass_ID, Charge_Termination);
	BQ27426_subclass_update(Discharge_SubClass_ID, Discharge);
	BQ27426_subclass_update(Registers_SubClass_ID, Registers);
	BQ27426_subclass_update(Power_SubClass_ID, Power);
	BQ27426_subclass_update(OCVa_0_SubClass_ID, OCVa_0);
	BQ27426_subclass_update(OCVa_1_SubClass_ID, OCVa_1);
	BQ27426_subclass_update(OCVb_0_SubClass_ID, OCVb_0);
	BQ27426_subclass_update(OCVb_1_SubClass_ID, OCVb_1);
	BQ27426_subclass_update(Resistance_Low_Temp_SubClass_ID, Resistance_Low_Temp);
	BQ27426_subclass_update(Resistance_High_Temp_SubClass_ID, Resistance_High_Temp);
#endif

	BQ27426_soft_reset();

	delay_ms(10);
}

static BQ27426_subclass_update(uint16_t SubClass_ID, const uint8_t *pData)
{
	uint8_t nData[32];
	uint8_t checksum, i;

	nData[0]=0x3E;
	nData[1]=((SubClass_ID >> 8) & 0xFF);
	nData[2]=(SubClass_ID & 0xFF);
	BQ27426_write(nData,3);

	BQ27426_read(0x60,nData,1);
	checksum = nData[0];
	checksum = 255 - checksum;

	BQ27426_read(0x40,nData,32);
	for(i=0; i<32; i++)
	{
		checksum = BQ27426_parameter_update(0x40+i, pData[i], checksum);
	}

	checksum = 255 - checksum;

	nData[0]=0x60; nData[1]=checksum;
	BQ27426_write(nData,2);
	delay_ms(10);

	BQ27426_read(0x40,nData,32);
	delay_ms(10);
}

static uint8_t BQ27426_parameter_update(uint8_t nregister, uint8_t byte, uint8_t checksum)
{
	uint8_t uData[2];

	BQ27426_read(nregister,uData,1);
	checksum = checksum - uData[0];

	uData[0]=nregister; uData[1]=byte;
	BQ27426_write(uData,2);
	checksum = checksum + uData[1];

	return checksum;
}

