/*
 * Copyright (c) 2016, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef RFID_RF430_H_
#define RFID_RF430_H_

#include "stdbool.h"
#include "stdint.h"
#include <inc/hw_memmap.h>

#include <ti/sysbios/knl/Task.h>
#include <ti/sysbios/knl/Clock.h>

/*********************************************************************
 * CONSTANTS and MACROS
 */

#define delay_ms(i) Task_sleep( ((i) * 1000) / Clock_tickPeriod )
#define MS_2_TICKS(ms) ( ((ms) * 1000) / Clock_tickPeriod )

#define RF430_DEV_ADDR		0x28

// define the registers
#define CONTROL_REG         0xFFFE
#define STATUS_REG          0xFFFC
#define INT_ENABLE_REG      0xFFFA
#define INT_FLAG_REG        0xFFF8
#define CRC_RESULT_REG      0xFFF6
#define CRC_LENGTH_REG      0xFFF4
#define CRC_START_ADDR_REG  0xFFF2
#define COMM_WD_CTRL_REG    0xFFF0
#define VERSION_REG         0xFFEE //contains the software version of the ROM
#define TEST_FUNCTION_REG   0xFFE2
#define TEST_MODE_REG       0xFFE0

//define the different virtual register bits
//CONTROL_REG bits
#define SW_RESET        0x0001
#define RF_ENABLE       0x0002
#define INT_ENABLE      0x0004
#define INTO_HIGH       0x0008
#define INTO_DRIVE      0x0010
#define BIP8_ENABLE     0x0020
#define STANDBY_ENABLE  0x0040
#define TEST430_ENABLE  0x0080

//STATUS_REG bits
#define READY           0x0001
#define CRC_ACTIVE      0x0002
#define RF_BUSY         0x0004

//INT_ENABLE_REG bits
#define EOR_INT_ENABLE      0x0002
#define EOW_INT_ENABLE      0x0004
#define CRC_INT_ENABLE      0x0008
#define BIP8_ERROR_INT_ENABLE       0x0010
#define NDEF_ERROR_INT_ENABLE       0x0020
#define GENERIC_ERROR_INT_ENABLE    0x0080

//INT_FLAG_REG bits
#define EOR_INT_FLAG    0x0002
#define EOW_INT_FLAG    0x0004
#define CRC_INT_FLAG    0x0008
#define BIP8_ERROR_INT_FLAG 0x0010
#define NDEF_ERROR_INT_FLAG 0x0020
#define GENERIC_ERROR_INT_FLAG  0x0080

//COMM_WD_CTRL_REG bits
#define WD_ENABLE   0x0001
#define TIMEOUT_PERIOD_2_SEC    0
#define TIMEOUT_PERIOD_32_SEC   0x0002
#define TIMEOUT_PERIOD_8_5_MIN  0x0004
#define TIMEOUT_PERIOD_MASK     0x0002 + 0x0004 + 0x0008

#define RF430_DEFAULT_DATA		{  														\
/*NDEF Tag Application Name*/ 															\
0xD2, 0x76, 0x00, 0x00, 0x85, 0x01, 0x01, 												\
																						\
/*Capability Container ID*/ 															\
0xE1, 0x03, 																			\
0x00, 0x0F,	/* CCLEN */																	\
0x20,		/* Mapping version 2.0 */													\
0x00, 0xF9,	/* MLe (49 bytes); Maximum R-APDU data size */								\
0x00, 0xF6, /* MLc (52 bytes); Maximum C-APDU data size */								\
0x04, 		/* Tag, File Control TLV (4 = NDEF file) */									\
0x06, 		/* Length, File Control TLV (6 = 6 bytes of data for this tag) */			\
0xE1, 0x04,	/* File Identifier */														\
0x0B, 0xDF, /* Max NDEF size (3037 bytes of useable memory) */							\
0x00, 		/* NDEF file read access condition, read access without any security */		\
0x00, 		/* NDEF file write access condition; write access without any security */	\
																						\
/* NDEF File ID */ 																		\
0xE1, 0x04, 																			\
																						\
/* NDEF File for Hello World */                             				            \
0x00, 0x14, /* NLEN: NDEF length (20 byte long message, max. length for RF430CL) */     \
                                                                                        \
/* NDEF Record (refer to NFC Data Exchange Format specifications)*/                     \
0xD1,       /*MB(Message Begin), SR(Short Record) flags set, ME(Message End), IL(ID length field present) flags cleared; TNF(3bits) = 1; */ \
0x01, 0x10, /*Type Length = 0x01; Payload Length = 0x10 */                              \
0x54,       /* Type = T (text) */                                                       \
0x02,       /* 1st payload byte: "Start of Text", as specified in ASCII Tables */       \
0x65, 0x6E, /* 'e', 'n', (2nd, 3rd payload bytes*/                                      \
                                                                                        \
/* 'Hello, world!' NDEF data*/                                                          \
'N', 'F', 'C', ' ', 'F', 'r', 'o', 'm', ' ', '2', '6', 'x', 'x'            \
	} /* End of data */

/*********************************************************************
 */

static void RF430_disableRF(void);
static void RF430_enableRF(void);

void RF430_configureOOBPairing(void);
void RF430_configureI2C(void);
void RF430_configureGPIO(void);

uint16_t RF430_readSingle(uint16_t ui16Address);
void RF430_writeSingle(uint16_t ui16Address, uint16_t ui16Value);
void RF430_writeContinuous(uint16_t ui16Address, uint8_t * ui8Data, uint16_t ui16Length);
void RF430_readContinuous(uint16_t ui16Address, uint8_t * ui8Data, uint16_t ui16Length);

void RF430_hardReset(void);

void MCU_delayuS(uint16_t ui16uSDelay);

#endif /* RFID_RF430_H_ */
